/**************************************************************************
 * BASF2 (Belle Analysis Framework 2)                                     *
 * Copyright(C) 2018 - Belle II Collaboration                             *
 *                                                                        *
 * Author: The Belle II Collaboration                                     *
 * Contributors: Felix Metzner                                            *
 *                                                                        *
 * This software is provided "as is" without any warranty.                *
 **************************************************************************/

#pragma once

#include <framework/core/Module.h>
#include <framework/datastore/StoreArray.h>
#include <framework/datastore/StoreObjPtr.h>

#include <mdst/dataobjects/Track.h>

#include <analysis/dataobjects/ParticleList.h>
#include <analysis/dataobjects/Particle.h>

#include <string>


namespace Belle2 {
  class Particle;

  /**
   * This module will loop over all tracks in an event and mark such tracks
     which fulfill the defined criteria to be considered as clone resulting
     from a curling track.
     The module evaluates pairs of tracks regarding their transverse (has to be
     lower than maxPt), the difference in their transverse momenta
     (abs(p_t1 - p_t2) < maxPtDiff) and the angle between their momenta
     (abs(cos(p1.Angle(p2)) > cos(maxAngle)).
     If a pair of tracks fulfilling those criteria is found, the one that does
     not have SVD hits is marked as the clone. If both or none have SVD hits,
     the track with the higher impact parameters ((d0/errd0)^2 + (z0/errz0)^2)
     is marked as clone.
   */
  class CurlerCloneMarkerModule : public Module {

  public:

    /// Constructor
    CurlerCloneMarkerModule();

    /// Module initialization.
    virtual void initialize() override;

    /// Event processor.
    virtual void event() override;

  private:
    /**
     * This function compares two tracks taking the given selection
     * criteria for curling tracks into account. If a pair of tracks
     * which might result from the same curling track is found, one
     * of them is marked as the clone of the curler.
     * The decision which of the two tracks is the clone is made based
     * on the SVD hits of the tracks, or, it this is not conclusive,
     * based on the impact parameters.
     */
    bool checkTrackPair(Track*, Track*);

    /// Module parameters
    double m_maxPt = 0.275;  /**< Maximal transverse momentum of a track to be considered as curler. */
    double m_maxPtDiff = 0.02;  /**< Maximal difference in trans. mom. of two tracks to result from the same curler. */
    double m_maxAngle = 5.0;  /**< Maximal angle between two track to result from the same cruler. */

    /// Member variables
    double m_maxAbsCosAngle;  /**< Cosine of m_maxAngle. */
  };
}
