"""
QED Correction Calculation
    Based on
    Impacts of radiative corrections on measurements of lepton flavour universality in B -> D l nu decays
    by Barbara Sciascia et al.
    https://arxiv.org/pdf/1905.02702.pdf

Felix Metzner 2019
"""

import numpy as np
import pandas as pd
from uncertainties import ufloat
from collections import OrderedDict

from typing import Union, Optional, Tuple, Sequence, Dict

from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.pdg_code_collections import PDGCodeCollection as PcC


__all__ = [
    "coulomb_correction_col_name",
    "new_coulomb_correction_col_name",
    "get_coulomb_corr_factor",
]


logger = log_to_default(name=__name__)


# Lepton Masses in GeV
m_e = ufloat(0.0005109989461, 0.0000000000031)  # http://pdglive.lbl.gov/DataBlock.action?node=S003M
m_mu = ufloat(0.1056583745, 0.0000000024)  # http://pdglive.lbl.gov/DataBlock.action?node=S004M
m_tau = ufloat(1.77686, 0.00012)  # http://pdglive.lbl.gov/DataBlock.action?node=S035M

# D Meson PDG Codes and Masses in GeV
d_meson_mass = OrderedDict(
    [
        # D+ http://pdglive.lbl.gov/DataBlock.action?node=S031M&home=MXXX035
        (PcC.d_charged_pdg_id, ufloat(1.86965, 0.00005)),
        # D*+ http://pdglive.lbl.gov/DataBlock.action?node=M062M&home=MXXX035
        (PcC.dst_charged_pdg_id, ufloat(2.01026, 0.00005)),
        # D_2*+ http://pdglive.lbl.gov/DataBlock.action?node=M150M&home=MXXX035
        (PcC.dtwostar_charged_pdg_id, ufloat(2.4654, 0.0013)),
        # D_0*+ http://pdglive.lbl.gov/DataBlock.action?node=M179M&home=MXXX035
        (PcC.dzerostar_charged_pdg_id, ufloat(2.349, 0.007)),
        # D_1+ http://pdglive.lbl.gov/DataBlock.action?node=M120M&home=MXXX035
        (PcC.done_charged_pdg_id, ufloat(2.4232, 0.0024)),
        # D'_1+ Using mass from Belle II evtgen evt.pdl file and self defined 0.01 uncert
        (PcC.dprimeone_charged_pdg_id, ufloat(2.445, 0.01)),
    ]
)


coulomb_correction_col_name: str = "__coulomb_corr_factor__"
new_coulomb_correction_col_name: str = "__new_coulomb_corr_factor__"


def get_coulomb_corr_factor(
    df: pd.DataFrame,
    col_name: Optional[str] = None,
    only_leptonic_taus: bool = False,
    d_mesons: Optional[Sequence[int]] = None,
    on_reco_sig_side: bool = False,
) -> None:
    if col_name is None:
        col_name = coulomb_correction_col_name

    if d_mesons is None:
        d_mesons = PcC.all_d_pdg_ids
    else:
        assert all(pdg in PcC.all_d_pdg_ids for pdg in d_mesons)
        assert all(pdg in d_meson_mass.keys() for pdg in d_mesons)

    assert col_name not in df.columns
    df.loc[:, col_name] = 1.0

    for index in [0, 1]:
        electrons = get_valid_b_mesons(
            df=df,
            index=index,
            lepton_pdg=11,
            d_mesons=tuple(d_mesons),
            on_reco_sig_side=on_reco_sig_side,
        )
        muons = get_valid_b_mesons(
            df=df,
            index=index,
            lepton_pdg=13,
            d_mesons=tuple(d_mesons),
            on_reco_sig_side=on_reco_sig_side,
        )
        taus = get_valid_b_mesons(
            df=df,
            index=index,
            lepton_pdg=15,
            d_mesons=tuple(d_mesons),
            on_reco_sig_side=on_reco_sig_side,
            only_leptonic_taus=only_leptonic_taus,
        )

        df.loc[electrons, col_name] *= calc_omega_c_corr_factor(
            df=df.loc[electrons],
            index=index,
            m_lepton=m_e.nominal_value,
        )
        df.loc[muons, col_name] *= calc_omega_c_corr_factor(
            df=df.loc[muons],
            index=index,
            m_lepton=m_mu.nominal_value,
        )
        df.loc[taus, col_name] *= calc_omega_c_corr_factor(
            df=df.loc[taus],
            index=index,
            m_lepton=m_tau.nominal_value,
        )

    assert not df[col_name].isnull().any()


def get_valid_b_mesons(
    df: pd.DataFrame,
    index: int,
    lepton_pdg: int,
    d_mesons: Tuple[int, ...],
    on_reco_sig_side: bool,
    only_leptonic_taus: bool = False,
) -> pd.Series:
    nu_pdg_mapping: Dict[int, int] = PcC.lepton_to_nu_pdg_mapping
    assert lepton_pdg in nu_pdg_mapping.keys(), lepton_pdg

    b_pdg_col = f"genUpsilon4S__bomcDaughter__bo{index}__cm__spPDG__bc__bc"
    d_pdg_col = f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"
    l_pdg_col = f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo1__cmPDG__bc__bc__bc"
    l_via_tau_pdg_col = (
        f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo1__cmmcDaughter__bo0__cmPDG__bc__bc__bc__bc"
    )
    nu_pdg_col = f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo2__cmPDG__bc__bc__bc"
    fourth_pdg_col = f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo3__cmPDG__bc__bc__bc"

    b_index_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomdstIndex__bc__bc__bc"
    b_via_tau_index_var = "daughter__bo1__cm__spdaughter__bo1__cm__spmcMother__bomcMother__bomdstIndex__bc__bc__bc__bc"

    # Check if neutral B and thus charged D meson
    is_neutral_b = abs(df[b_pdg_col]) == 511

    has_d_daughter = abs(df[d_pdg_col]).isin(d_mesons)
    has_right_lepton = abs(df[l_pdg_col]) == lepton_pdg
    has_right_neutrino = abs(df[nu_pdg_col]) == nu_pdg_mapping[lepton_pdg]
    has_no_fourth_daughter = df[fourth_pdg_col].isin(PcC.valid_forth_daughter)

    result = is_neutral_b & has_d_daughter & has_right_lepton & has_right_neutrino & has_no_fourth_daughter

    if only_leptonic_taus and lepton_pdg == 15:
        result = result & (abs(df[l_via_tau_pdg_col]).isin(PcC.light_lepton_pdg_ids))
    if on_reco_sig_side:
        if lepton_pdg == 15:
            result = result & (df[b_via_tau_index_var] == (index + 1))
        else:
            result = result & (df[b_index_var] == (index + 1))

    return result


def calc_omega_c_corr_factor(
    df: pd.DataFrame,
    index: int,
    m_lepton: Union[float, ufloat],
) -> np.ndarray:
    cols = {
        "d_px_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo0__cmpx__bc__bc__bc",
        "d_py_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo0__cmpy__bc__bc__bc",
        "d_pz_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo0__cmpz__bc__bc__bc",
        "d_e_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo0__cmE__bc__bc__bc",
        "l_px_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo1__cmpx__bc__bc__bc",
        "l_py_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo1__cmpy__bc__bc__bc",
        "l_pz_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo1__cmpz__bc__bc__bc",
        "l_e_col": f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo1__cmE__bc__bc__bc",
    }

    assert all(c in df.columns for c in cols.values())

    alpha: float = 1.0 / 137.0

    m_d_meson = get_meson_mass(df=df, index=index)

    m_l_sq = m_lepton * m_lepton
    m_d_sq = np.square(m_d_meson)

    s_dl = (
        +np.square(df[cols["d_e_col"]] + df[cols["l_e_col"]])
        - np.square(df[cols["d_px_col"]] + df[cols["l_px_col"]])
        - np.square(df[cols["d_py_col"]] + df[cols["l_py_col"]])
        - np.square(df[cols["d_pz_col"]] + df[cols["l_pz_col"]])
    )

    beta_dl = np.sqrt(1.0 - (4.0 * m_d_sq * m_l_sq) / np.square(s_dl - m_d_sq - m_l_sq))

    x_c = 2.0 * np.pi * alpha / beta_dl
    omega_c = x_c / (1.0 - np.exp(-x_c))

    # Reset cases where argument of sqrt in beta_dl is negative to 1.0:
    if any(np.isnan(omega_c)):
        logger.warning(
            f"Encountered {np.sum(np.isnan(omega_c))} events of {omega_c.shape[0]} with unphysical values! "
            f"This is probably due to negative values in the square root calculation of beta_dl. "
            f"For these events with the lepton mass {m_lepton}, D meson mass {m_d_meson} "
            f"and B meson mdst index {index} the weight is set to 1.0."
        )
        omega_c[np.isnan(omega_c)] = 1.0

    assert np.all(omega_c > 0.0), (m_lepton, index, np.sum(omega_c <= 0), omega_c.shape[0])

    return omega_c


def get_meson_mass(
    df: pd.DataFrame,
    index: int,
) -> np.ndarray:
    d_pdg_col = f"genUpsilon4S__bomcDaughter__bo{index}__cm__spmcDaughter__bo0__cmPDG__bc__bc__bc"

    conditions = [abs(df[d_pdg_col]) == pdg for pdg in d_meson_mass.keys()]
    choices = [mass.nominal_value for mass in d_meson_mass.values()]
    result = np.select(condlist=conditions, choicelist=choices, default=np.nan)
    assert not np.isnan(result).any()
    return result
