from typing import Optional, Dict, Tuple

from rdstar.offline_analysis.fitting.fit_case import FitCase, FitVariable

from rdstar.offline_analysis.dss_fitting.utilities import dss_ff_study_utils
from rdstar.offline_analysis.dss_fitting.utilities import utilities as fitting_utils

from templatefitter.binned_distributions.binning import BinsInputType


__all__ = [
    "FitCaseDict",
    # Asimov
    "reduced_dss_asimov_fit_case_dict_for_DPiZ",
    "reduced_dss_asimov_fit_case_dict_for_DSPiZ",
    "reduced_dss_asimov_fit_case_dict_for_DPiP",
    "reduced_dss_asimov_fit_case_dict_for_DSPiP",
    "alternative_dss_asimov_fit_case_dict_for_DPiZ",
    "alternative_dss_asimov_fit_case_dict_for_DSPiZ",
    "alternative_dss_asimov_fit_case_dict_for_DPiP",
    "alternative_dss_asimov_fit_case_dict_for_DSPiP",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DPiP",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiP",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DPiZ",
    "dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiZ",
    "dss_pStarL_alternative_asimov_fit_case_dict_for_DPiP",
    "dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiP",
    "dss_pStarL_alternative_asimov_fit_case_dict_for_DPiZ",
    "dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiZ",
    # Data
]


BinningDictType = Tuple[Optional[Dict[str, BinsInputType]], ...]
ScopeDictType = Tuple[Optional[Dict[str, Tuple[float, float]]], ...]


class FitCaseDict:
    def __init__(
        self,
        pivot_fit_case: FitCase,
        special_fit_variables_dict_per_bin: Optional[Dict[int, Dict[str, Tuple[FitVariable, ...]]]] = None,
    ) -> None:

        self._pivot_fit_case = pivot_fit_case
        self._special_fit_variables_dict_per_bin = special_fit_variables_dict_per_bin

    def get_fit_variables_dict_in_bin(
        self,
        bin_number: int,
    ) -> Dict[str, Tuple[FitVariable, ...]]:
        assert self._special_fit_variables_dict_per_bin is not None
        assert bin_number in self._special_fit_variables_dict_per_bin.keys()
        return self._special_fit_variables_dict_per_bin[bin_number]

    def get_fit_case_in_bin(
        self,
        bin_number: int,
    ) -> FitCase:
        if not self._special_fit_variables_dict_per_bin:
            return self._pivot_fit_case
        elif bin_number not in self._special_fit_variables_dict_per_bin.keys():
            return self._pivot_fit_case
        else:
            fit_variables_dict = self.get_fit_variables_dict_in_bin(bin_number=bin_number)
            return FitCase(
                name=self.pivot_fit_case.name,
                base_dir_name=self.pivot_fit_case.base_dir_name,
                sample_key=self.pivot_fit_case.sample_key,
                output_tag=self.pivot_fit_case.output_tag,
                plot_info_text=self.pivot_fit_case.plot_info_text,
                fix_nuisance_parameters=self.pivot_fit_case.fix_nuisance_parameters,
                asimov_data=self.pivot_fit_case.asimov_data,
                round_asimov_data=self.pivot_fit_case.round_asimov_data,
                fit_variables_by_mode_dict=fit_variables_dict,
                weight_column_name=self.pivot_fit_case.weight_column_name,
                systematics_cols=self.pivot_fit_case.systematics_cols,
                reco_mode_infos=self.pivot_fit_case.reco_mode_infos,
                template_setup=self.pivot_fit_case.template_setup,
                sig_id_to_data_type_mapping=self.pivot_fit_case.sig_id_to_data_type_mapping,
                plot_colors=self.pivot_fit_case.plot_colors,
                plot_labels=self.pivot_fit_case.plot_labels,
                fix_component_yields=self.pivot_fit_case.fix_component_yields,
                sig_id_column_name=self.pivot_fit_case.sig_id_column_name,
                sig_queries=self.pivot_fit_case.sig_queries,
                reco_mode_labels_dict=self.pivot_fit_case.reco_mode_labels_dict,
                alternative_template_colors=self.pivot_fit_case.alternative_template_colors,
                minimizer_id=self.pivot_fit_case.minimizer_id,
                create_plots=self.pivot_fit_case.create_plots,
                create_projection_plots=self.pivot_fit_case.create_projection_plots,
                create_template_plots=self.pivot_fit_case.create_template_plots,
                create_2d_template_plots=self.pivot_fit_case.create_2d_template_plots,
                additional_info=self.pivot_fit_case.additional_info,
            )

    @property
    def pivot_fit_case(self) -> FitCase:
        return self._pivot_fit_case

    @property
    def name(self) -> str:
        return self.pivot_fit_case.name


# Asimov Fit Cases and Dicts
# q^2
reduced_dss_asimov_fit_case_for_DPiZ = FitCase(
    name="reduced_dss_asimov_for_DPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DPiZ = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DPiZ,
    special_fit_variables_dict_per_bin={
        i: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm] for rm in fitting_utils.dss_modes_for_DPiZ.keys()
        }
        for i in range(0, 3)
    },
)

reduced_dss_asimov_fit_case_for_DSPiZ = FitCase(
    name="reduced_dss_asimov_for_DSPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DSPiZ = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DSPiZ,
    special_fit_variables_dict_per_bin={
        i: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm]
            for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
        }
        for i in range(0, 3)
    },
)

reduced_dss_asimov_fit_case_for_DPiP = FitCase(
    name="reduced_dss_asimov_for_DPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DPiP = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DPiP,
    special_fit_variables_dict_per_bin={
        i: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm] for rm in fitting_utils.dss_modes_for_DPiP.keys()
        }
        for i in range(0, 3)
    },
)

reduced_dss_asimov_fit_case_for_DSPiP = FitCase(
    name="reduced_dss_asimov_for_DSPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DSPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
reduced_dss_asimov_fit_case_dict_for_DSPiP = FitCaseDict(
    pivot_fit_case=reduced_dss_asimov_fit_case_for_DSPiP,
    special_fit_variables_dict_per_bin={
        i: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm]
            for rm in fitting_utils.dss_modes_for_DSPiP.keys()
        }
        for i in range(0, 3)
    },
)

alternative_dss_asimov_fit_case_for_DPiZ = FitCase(
    name="alternative_dss_asimov_for_DPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
alternative_dss_asimov_fit_case_dict_for_DPiZ = FitCaseDict(
    pivot_fit_case=alternative_dss_asimov_fit_case_for_DPiZ,
    special_fit_variables_dict_per_bin=None,
)

alternative_dss_asimov_fit_case_for_DSPiZ = FitCase(
    name="alternative_dss_asimov_for_DSPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
alternative_dss_asimov_fit_case_dict_for_DSPiZ = FitCaseDict(
    pivot_fit_case=alternative_dss_asimov_fit_case_for_DSPiZ,
    special_fit_variables_dict_per_bin=None,
)

alternative_dss_asimov_fit_case_for_DPiP = FitCase(
    name="alternative_dss_asimov_for_DPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
alternative_dss_asimov_fit_case_dict_for_DPiP = FitCaseDict(
    pivot_fit_case=alternative_dss_asimov_fit_case_for_DPiP,
    special_fit_variables_dict_per_bin=None,
)

alternative_dss_asimov_fit_case_for_DSPiP = FitCase(
    name="alternative_dss_asimov_for_DSPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DSPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
alternative_dss_asimov_fit_case_dict_for_DSPiP = FitCaseDict(
    pivot_fit_case=alternative_dss_asimov_fit_case_for_DSPiP,
    special_fit_variables_dict_per_bin=None,
)

# pStarL
pStarL_reduced_dss_asimov_fit_case_for_DPiZ = FitCase(
    name="pStarL_reduced_dss_asimov_for_DPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DPiZ = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DPiZ,
    special_fit_variables_dict_per_bin={
        5: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm] for rm in fitting_utils.dss_modes_for_DPiZ.keys()
        },
        6: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm] for rm in fitting_utils.dss_modes_for_DPiZ.keys()
        },
        7: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm] for rm in fitting_utils.dss_modes_for_DPiZ.keys()
        },
    },
)

pStarL_reduced_dss_asimov_fit_case_for_DSPiZ = FitCase(
    name="pStarL_reduced_dss_asimov_for_DSPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiZ = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DSPiZ,
    special_fit_variables_dict_per_bin={
        5: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm]
            for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
        },
        6: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm]
            for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
        },
        7: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm]
            for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
        },
    },
)

pStarL_reduced_dss_asimov_fit_case_for_DPiP = FitCase(
    name="pStarL_reduced_dss_asimov_for_DPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DPiP = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DPiP,
    special_fit_variables_dict_per_bin={
        5: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm] for rm in fitting_utils.dss_modes_for_DPiP.keys()
        },
        6: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm] for rm in fitting_utils.dss_modes_for_DPiP.keys()
        },
        7: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm] for rm in fitting_utils.dss_modes_for_DPiP.keys()
        },
    },
)

pStarL_reduced_dss_asimov_fit_case_for_DSPiP = FitCase(
    name="pStarL_reduced_dss_asimov_for_DSPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.ff_study_fit_variables_dict[rm] for rm in fitting_utils.dss_modes_for_DSPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_reduced_asimov_fit_case_dict_for_DSPiP = FitCaseDict(
    pivot_fit_case=pStarL_reduced_dss_asimov_fit_case_for_DSPiP,
    special_fit_variables_dict_per_bin={
        5: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_1[rm]
            for rm in fitting_utils.dss_modes_for_DSPiP.keys()
        },
        6: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm]
            for rm in fitting_utils.dss_modes_for_DSPiP.keys()
        },
        7: {
            rm: dss_ff_study_utils.ff_study_fit_variables_dict_opt_2[rm]
            for rm in fitting_utils.dss_modes_for_DSPiP.keys()
        },
    },
)

pStarL_alternative_dss_asimov_fit_case_for_DPiZ = FitCase(
    name="pStarL_alternative_dss_asimov_fit_case_for_DPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_pStarL_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_alternative_asimov_fit_case_dict_for_DPiZ = FitCaseDict(
    pivot_fit_case=pStarL_alternative_dss_asimov_fit_case_for_DPiZ,
    special_fit_variables_dict_per_bin=None,
)

pStarL_alternative_dss_asimov_fit_case_for_DSPiZ = FitCase(
    name="pStarL_alternative_dss_asimov_fit_case_for_DSPiZ",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_pStarL_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DSPiZ.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiZ.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiZ,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiZ = FitCaseDict(
    pivot_fit_case=pStarL_alternative_dss_asimov_fit_case_for_DSPiZ,
    special_fit_variables_dict_per_bin=None,
)

pStarL_alternative_dss_asimov_fit_case_for_DPiP = FitCase(
    name="pStarL_alternative_dss_asimov_fit_case_for_DPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_pStarL_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_alternative_asimov_fit_case_dict_for_DPiP = FitCaseDict(
    pivot_fit_case=pStarL_alternative_dss_asimov_fit_case_for_DPiP,
    special_fit_variables_dict_per_bin=None,
)

pStarL_alternative_dss_asimov_fit_case_for_DSPiP = FitCase(
    name="pStarL_alternative_dss_asimov_fit_case_for_DSPiP",
    base_dir_name="Fitting_Asimov_DSS",
    sample_key="",
    output_tag="AsimovFit_DSS_FixedNuisanceParams",
    plot_info_text="Asimov Data",
    fix_nuisance_parameters=True,
    asimov_data=True,
    round_asimov_data=False,
    fit_variables_by_mode_dict={
        rm: dss_ff_study_utils.alternative_ff_study_pStarL_fit_variables_dict[rm]
        for rm in fitting_utils.dss_modes_for_DSPiP.keys()
    },
    weight_column_name=fitting_utils.final_weight_column_for_asimov,
    systematics_cols=tuple([]),
    reco_mode_infos=[(k, v) for k, v in fitting_utils.dss_modes_for_DSPiP.items()],
    reco_mode_labels_dict=fitting_utils.dss_labels,
    template_setup=dss_ff_study_utils.dss_ff_template_setup_for_DSPiP,
    sig_id_to_data_type_mapping={},
    plot_labels=dss_ff_study_utils.reduced_dss_labels,
    plot_colors=dss_ff_study_utils.reduced_dss_colors,
    sig_id_column_name="advanced_sig_id",
    sig_queries=dss_ff_study_utils.advanced_dss_components,
)
dss_pStarL_alternative_asimov_fit_case_dict_for_DSPiP = FitCaseDict(
    pivot_fit_case=pStarL_alternative_dss_asimov_fit_case_for_DSPiP,
    special_fit_variables_dict_per_bin=None,
)
