"""
Tools for running and evaluating the continuum cal fits

Felix Metzner 2023
"""

import os
import copy
import json
import numpy as np
import pandas as pd

from uncertainties import unumpy as unp

from iminuit import Minuit
from iminuit.minuit import mutil
from typing import Tuple, Dict, Type, Any, Optional

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_units import LatexUnits
from rdstar.latex_labels.latex_observable_names import LatexObsNames

from rdstar.comparison_tools.select_utilities import SimpleCutInfo
from rdstar.comparison_tools.data_mc_comparer import DataMCComparer

from rdstar.offline_analysis.systematics import full_systematics_cols
from rdstar.offline_analysis.histogram_variables import RDStarHistVars
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.rdstar_samples import alternative_sample_keys
from rdstar.offline_analysis.validation.sideband_validation import SidebandInfo

from rdstar.offline_analysis.reco_modes.reco_mode_infos import TagRecoModeInfos, U4SRecoModeInfos
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BTagRecoModes, BSigRecoModes, BtoDssLRecoModes

from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import get_file_path_for

from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_evaluation import apply_fei_calibration
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import right_sig_tag_combination_selection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters.cont_cal_fitter_base import ContCalFitterBase

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.cont_cal_third_approach_fit_info import (
    continuum_cal_fit__component_info_collection,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.cont_cal_in_dss_1st_approach_fit_info import (
    continuum_cal_for_dss_fit__component_info_collection,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_fitters import (
    cont_cal_fitter_dict,
    keys_to_evaluate,
    n_e_ecl_extra_selection,
)


__all__ = [
    "ContCalEvalResult",
    "ContCalFitEvaluator",
]


logger = log_to_default(name=__name__)


# region sideband eval infos

sideband_data_component_label: str = "Sideband Data"
cs_mva_sideband_title: str = rf"{LatexObsNames.mva_cs_output} Sideband"
_cs_mva_sideband_text: str = r"0.2 < $" + LatexObsNames.mva_cs_output[1:-1] + r" < 0.85$"
_cs_mva_sideband_text += "\n" + rf"{LatexObsNames.roe_n_e_ecl_extra[:-1]} > 0.75\,{LatexUnits.gev[1:]}"
cs_mva_sideband_text: str = _cs_mva_sideband_text

post_fit_plot_reco_mode_info_collection = BSigRecoModes.b_to_dwwos_l_modes_by_b_charge
post_fit_plot_reco_mode_info_collection_for_dss = BtoDssLRecoModes.b_to_dss_l_modes_by_d_charge
# post_fit_plot_mc_matching_collection = MCMatchingCollections.extended_sig_id__basic__collection
post_fit_plot_mc_matching_collection = continuum_cal_fit__component_info_collection
post_fit_plot_mc_matching_collection_for_dss = continuum_cal_for_dss_fit__component_info_collection

eval_sb_plot_infos: Tuple[SidebandInfo, ...] = (
    SidebandInfo(
        key="cs_mva_sideband",
        hist_var=RDStarHistVars.cs_cos_tbto_cs_sideband_var.get_modified_histogram_variable(n_bins=20),
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 1),
        y_scale=(1.3, 1.9),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
    SidebandInfo(
        key="cs_mva_sideband",
        hist_var=RDStarHistVars.missing_mass_alt_res_corr_q2sb_wide,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.97),
        legend_loc=(1, 2),
        legend_cols=(1, 1),
        y_scale=(1.3, 2.0),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
    SidebandInfo(
        key="cs_mva_sideband",
        hist_var=RDStarHistVars.pl_star_signal_fit,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.70),
        text_v_pos=(0.97, 0.68),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.9),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
    SidebandInfo(
        key="cs_mva_sideband",
        hist_var=RDStarHistVars.neextra_for_fit,
        title=cs_mva_sideband_title,
        info_text=cs_mva_sideband_text,
        reco_mode_info_collection=post_fit_plot_reco_mode_info_collection,
        mc_matching_collection=post_fit_plot_mc_matching_collection,
        text_h_pos=(0.03, 0.68),
        text_v_pos=(0.97, 0.76),
        legend_loc=(1, 2),
        legend_cols=(1, 2),
        y_scale=(1.3, 1.9),
        systematics=None,
        proportional_rescaling_for_all_modes_case=False,
        data_component_label=sideband_data_component_label,
        file_name_appendix="post_fit_plot",
    ),
)

# endregion


# region ContCalEvalResult
class NumpyNDArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class ContCalEvalResult:
    def __init__(
        self,
        selection_name: str,
        fit_scenario_name: str,
        fit_obs_setup_tag: str,
        initial_parameters: np.ndarray,
        cov_mat: np.ndarray,
        corr_mat: np.ndarray,
        res_values: np.ndarray,
        res_errors: np.ndarray,
        param_names_mapping: Dict[int, str],
        component_mapping: Dict[int, str],
    ) -> None:
        self.selection_name: str = selection_name
        self.fit_scenario_name: str = fit_scenario_name
        self.fit_obs_setup_tag: str = fit_obs_setup_tag

        self.initial_parameters: np.ndarray = initial_parameters

        self.cov_mat: np.ndarray = cov_mat
        self.corr_mat: np.ndarray = corr_mat
        self.res_values: np.ndarray = res_values
        self.res_errors: np.ndarray = res_errors
        self.param_names_mapping: Dict[int, str] = param_names_mapping
        self.component_mapping: Dict[int, str] = component_mapping

        self.inverted_param_names_mapping: Dict[str, int] = {v: i for i, v in self.param_names_mapping.items()}
        self.inverted_component_mapping: Dict[str, int] = {v: i for i, v in self.component_mapping.items()}

    @property
    def data_over_mc_ratio_array(self) -> unp.uarray:
        fit_res_array = unp.uarray(nominal_values=self.res_values, std_devs=self.res_errors)
        return fit_res_array / self.initial_parameters

    def get_latex_table_col_str_array(self, precision: int) -> np.ndarray:
        return np.array([f"${_v:.{precision}L}$" for _v in self.data_over_mc_ratio_array])

    def get_latex_table_col_str(self, precision: int) -> str:
        return "\n".join([f"${_v:.{precision}L}$" for _v in self.data_over_mc_ratio_array])

    @property
    def params_names_col_str(self) -> str:
        return "\n".join(self.param_names_mapping.values())

    @property
    def params_names_latex_labels_array(self) -> Tuple[str, ...]:
        cont_cal_fit_setup = self.fit_scenario.this_fitter_setup
        assert cont_cal_fit_setup is not None
        return tuple(cont_cal_fit_setup.asimov.components[i].latex_label for i in self.param_names_mapping.keys())

    def as_dict(self) -> Dict[str, Any]:
        return {
            "selection_name": self.selection_name,
            "fit_scenario_name": self.fit_scenario_name,
            "fit_obs_setup_tag": self.fit_obs_setup_tag,
            "initial_parameters": self.initial_parameters,
            "cov_mat": self.cov_mat,
            "corr_mat": self.corr_mat,
            "res_values": self.res_values,
            "res_errors": self.res_errors,
            "inverted_param_names_mapping": self.inverted_param_names_mapping,
            "inverted_component_mapping": self.inverted_component_mapping,
        }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "ContCalEvalResult":
        _inverted_param_names_mapping = input_dict["inverted_param_names_mapping"]
        assert isinstance(_inverted_param_names_mapping, dict), type(_inverted_param_names_mapping).__name__
        _inverted_component_mapping = input_dict["inverted_component_mapping"]
        assert isinstance(_inverted_component_mapping, dict), type(_inverted_component_mapping).__name__
        _param_names_mapping: Dict[int, str] = {int(v): i for i, v in _inverted_param_names_mapping.items()}
        _component_mapping: Dict[int, str] = {int(v): i for i, v in _inverted_component_mapping.items()}

        return ContCalEvalResult(
            selection_name=input_dict["selection_name"],
            fit_scenario_name=input_dict["fit_scenario_name"],
            fit_obs_setup_tag=input_dict["fit_obs_setup_tag"],
            initial_parameters=np.asarray(input_dict["initial_parameters"]),
            cov_mat=np.asarray(input_dict["cov_mat"]),
            corr_mat=np.asarray(input_dict["corr_mat"]),
            res_values=np.asarray(input_dict["res_values"]),
            res_errors=np.asarray(input_dict["res_errors"]),
            param_names_mapping=_param_names_mapping,
            component_mapping=_component_mapping,
        )

    def save_to_file(self, file_path: PathType) -> None:
        with open(file_path, "w") as fp:
            json.dump(obj=self.as_dict(), fp=fp, indent=2, cls=NumpyNDArrayEncoder)

    @staticmethod
    def load_from_file(file_path: PathType) -> "ContCalEvalResult":
        with open(file_path) as f:
            _loaded_dict = json.load(f)
        assert isinstance(_loaded_dict, dict), type(_loaded_dict).__name__
        return ContCalEvalResult.from_dict(input_dict=_loaded_dict)

    @property
    def fit_scenario(self) -> Type[ContCalFitterBase]:
        return cont_cal_fitter_dict[self.fit_scenario_name]

    @staticmethod
    def compare_results(
        dir_path: PathType,
        first_eval_name: str,
        first_scenario_name: str,
        first_selection_name: str,
        first_obs_setup_tag: str,
        second_eval_name: str,
        second_scenario_name: str,
        second_selection_name: str,
        second_obs_setup_tag: str,
        precision: int = 2,
    ) -> pd.DataFrame:
        first_file_name: str = f"{first_scenario_name}__{first_selection_name}__{first_obs_setup_tag}.json"
        second_file_name: str = f"{second_scenario_name}__{second_selection_name}__{second_obs_setup_tag}.json"

        first_file_path: PathType = os.path.join(dir_path, first_eval_name, first_file_name)
        second_file_path: PathType = os.path.join(dir_path, second_eval_name, second_file_name)

        assert os.path.isfile(first_file_path), first_file_path
        assert os.path.isfile(second_file_path), second_file_path

        first_result_info = ContCalEvalResult.load_from_file(file_path=first_file_path)
        second_result_info = ContCalEvalResult.load_from_file(file_path=second_file_path)

        first_df = pd.DataFrame(
            {
                first_scenario_name: first_result_info.params_names_latex_labels_array,
                first_selection_name: first_result_info.get_latex_table_col_str_array(precision=precision),
            }
        )
        second_df = pd.DataFrame(
            {
                second_scenario_name: second_result_info.params_names_latex_labels_array,
                second_selection_name: second_result_info.get_latex_table_col_str_array(precision=precision),
            }
        )
        return pd.concat([first_df, second_df], axis=1)

    def get_mc_sample_with_correction(
        self,
        df: pd.DataFrame,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        _sample = copy.deepcopy(df)

        assert self.fit_scenario.this_fitter_setup is not None
        reco_id_col: str = self.fit_scenario.this_fitter_setup.asimov.reco_mode_col
        component_id_col: str = self.fit_scenario.this_fitter_setup.asimov.component_id_column_name
        weight_col: str = self.fit_scenario.this_fitter_setup.asimov.weight_info.col_name

        if component_id_col not in _sample.columns:
            self.fit_scenario.this_fitter_setup.asimov.prepare_mc_df(df=_sample)

        if matching_tag_selection is None:
            tag_match_mask: np.ndarray = np.full(shape=len(_sample), fill_value=True)
        elif matching_tag_selection is True:
            tag_match_mask = _sample[Cols.u4s_reco_mode_column_name].isin(U4SRecoModeInfos.u4s_from_all_correct_modes.ids)
        elif matching_tag_selection is False:
            tag_match_mask = _sample[Cols.u4s_reco_mode_column_name].isin(
                U4SRecoModeInfos.u4s_from_all_wrong_tag_modes.ids
            )
        else:
            raise RuntimeError(f"Invalid option 'matching_tag_selection' = {matching_tag_selection}!")

        test_mask: np.ndarray = np.full(shape=len(_sample), fill_value=False)

        for reco_ids, ratio_index in self.fit_scenario.get_correction_mapping().items():
            reco_mask: np.ndarray = _sample[reco_id_col].isin(reco_ids)
            is_continuum: np.ndarray = np.equal(_sample[Cols.is_continuum_evt].values, np.ones(shape=reco_mask.shape))

            mask: np.ndarray = reco_mask & is_continuum & tag_match_mask

            scale_factor: float = self.data_over_mc_ratio_array[ratio_index].nominal_value
            _sample.loc[mask, weight_col] *= scale_factor

            assert not np.any(test_mask & mask), np.sum(test_mask & mask)
            test_mask |= mask

        return _sample

    def get_mc_sample_with_simple_correction(
        self,
        df: pd.DataFrame,
        matching_tag_selection: Optional[bool] = None,
    ) -> pd.DataFrame:
        raise NotImplementedError("Method get_mc_sample_with_simple_correction is not available for ContCalEvalResult!")


# endregion


# region ContCalFitEvaluator
class ContCalFitEvaluator:

    selections: Tuple[str, ...] = (
        # "main_selection",
        "main_with_mva_selection_sec_sel_mva",
        # "first_selection",
    )

    u4s_combinations_selection: Dict[str, Optional[SimpleCutInfo]] = {
        # "without_u4s_selection": None,
        "with_right_sig_tag_comb": right_sig_tag_combination_selection,
    }

    sb_plot_infos: Tuple[SidebandInfo, ...] = eval_sb_plot_infos

    def __init__(
        self,
        evaluation_name: str,
        base_input_path: PathType,
        base_output_path: PathType,
        fit_obs_setup_tag: str,
        produce_plots: bool = True,
    ) -> None:

        assert os.path.isdir(base_input_path), base_input_path
        self.evaluation_name: str = evaluation_name
        self.base_input_path: PathType = base_input_path
        self.base_output_path: PathType = base_output_path
        self.fit_obs_setup_tag: str = fit_obs_setup_tag
        self.produce_plots: bool = produce_plots

        self._mc_sample: Optional[pd.DataFrame] = None

        self.output_path: PathType = os.path.join(self.base_output_path, self.evaluation_name)

        self.fit_scenario_dict: Dict[str, Type[ContCalFitterBase]] = copy.deepcopy(cont_cal_fitter_dict)

    @property
    def mc_sample(self) -> pd.DataFrame:
        assert self._mc_sample is not None
        return self._mc_sample

    @mc_sample.setter
    def mc_sample(self, df: pd.DataFrame) -> None:
        assert self._mc_sample is None
        self._mc_sample = df

    def reset_mc_sample(self, df: pd.DataFrame) -> None:
        self._mc_sample = df

    def run_evaluation(self) -> None:
        assert os.path.isdir(self.base_output_path), self.base_output_path
        os.makedirs(self.output_path, exist_ok=True)

        logger.info("Starting evaluation!")
        for selection_name in self.selections:
            logger.info(f"\tEvaluating selection {selection_name}")
            self.evaluate_selection(selection_name=selection_name)

    def evaluate_selection(
        self,
        selection_name: str,
    ) -> None:
        logger.info("\t\tloading mc data...")
        self.reset_mc_sample(df=self.load_mc(selection_name=selection_name))
        logger.info("\t\tloading recorded data...")
        data_sample: pd.DataFrame = self.load_data(selection_name=selection_name)

        for fit_scenario_name in self.fit_scenario_dict:
            if keys_to_evaluate and fit_scenario_name not in keys_to_evaluate:
                continue

            for u4s_sel_name, u4s_sel in self.u4s_combinations_selection.items():
                logger.info(f"\t\t- fit_scenario {fit_scenario_name} {u4s_sel_name} Y(4S) Selection")

                _u4s_mc_sample: pd.DataFrame = copy.deepcopy(self.mc_sample)
                _u4s_data_sample: pd.DataFrame = copy.deepcopy(data_sample)

                if u4s_sel is not None:
                    u4s_sel.apply_cut(df=_u4s_mc_sample)
                    u4s_sel.apply_cut(df=_u4s_data_sample)

                    _u4s_mc_sample = _u4s_mc_sample.reset_index(drop=True)
                    _u4s_data_sample = _u4s_data_sample.reset_index(drop=True)

                self.evaluate_scenario(
                    mc_sample_copy=_u4s_mc_sample,
                    data_sample=_u4s_data_sample,
                    selection_name=f"{selection_name}_{u4s_sel_name}",
                    fit_scenario_name=fit_scenario_name,
                )

    def evaluate_scenario(
        self,
        mc_sample_copy: pd.DataFrame,
        data_sample: pd.DataFrame,
        selection_name: str,
        fit_scenario_name: str,
    ) -> None:
        cont_cal_fitter: Type[ContCalFitterBase] = self.fit_scenario_dict[fit_scenario_name]
        assert cont_cal_fitter.this_fitter_setup is not None
        fit_scenario_triplet: FitSetupTriplet = cont_cal_fitter.this_fitter_setup
        data_fit_setup: FitSetupInfoContainer = fit_scenario_triplet.data

        logger.info("\t\t\tpreparing mc sample...")
        if data_fit_setup.component_id_column_name not in mc_sample_copy.columns:
            data_fit_setup.prepare_mc_df(df=mc_sample_copy)

        logger.info("\t\t\tsetting up fit...")
        fitter: ContCalFitterBase = cont_cal_fitter(
            fit_setup_info=data_fit_setup,
            mc_df=mc_sample_copy,
            data_df=data_sample,
        )

        initial_parameters: np.ndarray = np.array(list(fitter.initial_parameters))

        logger.info("\t\t\tfitting...")
        m = Minuit(
            fitter.cost_fcn,
            initial_parameters,
            name=tuple([str(i) for i in range(len(list(fitter.initial_parameters)))]),
        )

        fitter.set_iminuit_properties(minuit_instance=m)

        data_fmin: mutil.FMin = (
            m.simplex()
            .migrad(ncall=fitter.minuit_settings.n_call, iterate=fitter.minuit_settings.n_iterate)
            .minos()
            .hesse()
            .fmin
        )

        assert data_fmin.is_valid and data_fmin.has_valid_parameters and data_fmin.has_covariance
        logger.info("\t\t\tDone! Saving now...")

        eval_result: ContCalEvalResult = self.generate_result_output(
            selection_name=selection_name,
            fit_scenario_name=fit_scenario_name,
            fitter=fitter,
            minimizer=m,
            initial_parameters=initial_parameters,
        )

        self.save_results(
            selection_name=selection_name,
            fit_scenario_name=fit_scenario_name,
            result=eval_result,
        )

        if self.produce_plots:
            logger.info("\t\t\tProducing post fit plots...")
            self.produce_post_fit_plots(
                selection_name=selection_name,
                fit_scenario_name=fit_scenario_name,
                mc_sample=mc_sample_copy,
                data_sample=data_sample,
                fitter=fitter,
                minimizer=m,
            )
        else:
            logger.info("\t\t\tSkipping post fit plots!")

    def generate_result_output(
        self,
        selection_name: str,
        fit_scenario_name: str,
        fitter: ContCalFitterBase,
        minimizer: Minuit,
        initial_parameters: np.ndarray,
    ) -> ContCalEvalResult:

        assert np.all(initial_parameters == fitter.initial_parameters), (initial_parameters, fitter.initial_parameters)

        return ContCalEvalResult(
            selection_name=selection_name,
            fit_scenario_name=fit_scenario_name,
            fit_obs_setup_tag=self.fit_obs_setup_tag,
            initial_parameters=initial_parameters,
            cov_mat=np.array(minimizer.covariance),
            corr_mat=np.array(minimizer.covariance.correlation()),
            res_values=np.array(minimizer.values),
            res_errors=np.array(minimizer.errors),
            param_names_mapping=fitter.parameter_name_mapping(),
            component_mapping=fitter.fit_setup_info.components_name_to_id_mapping,
        )

    def produce_post_fit_plots(
        self,
        selection_name: str,
        fit_scenario_name: str,
        mc_sample: pd.DataFrame,
        data_sample: pd.DataFrame,
        fitter: ContCalFitterBase,
        minimizer: Minuit,
    ) -> None:
        output_name: str = f"{fit_scenario_name}__{selection_name}__{self.fit_obs_setup_tag}"
        output_path: PathType = os.path.join(self.output_path, output_name)

        if os.path.isdir(output_path):
            return

        os.makedirs(output_path, exist_ok=True)

        rescaled_mc_sample = fitter.get_rescaled_mc_sample_with(
            sample=mc_sample,
            weight_col_name=fitter.mc_weight_col_name,
            parameters=np.array(minimizer.values),
            add_required_sig_id_col=False,
        )

        for _sb_info in self.sb_plot_infos:
            if fit_scenario_name == "cont_cal_main_for_dss":
                sb_info = SidebandInfo.create_modified_comparer_info(
                    comparer_info=_sb_info,
                    mc_matching_collection=post_fit_plot_mc_matching_collection_for_dss,
                    reco_mode_info_collection=post_fit_plot_reco_mode_info_collection_for_dss,
                )
            else:
                sb_info = _sb_info

            sb_plot_output_path: PathType = os.path.join(output_path, sb_info.hist_var.df_label)
            os.makedirs(sb_plot_output_path, exist_ok=True)

            for use_sys_for_plot in (True, False):
                if use_sys_for_plot:
                    sb_info_with_sys = SidebandInfo.create_comparer_infos_with_systematics(
                        comparer_info_list=[sb_info],
                        systematics_collection=full_systematics_cols,
                        sys_name_tag="with_sys",
                    )[0]
                else:
                    sb_info_with_sys = sb_info

                use_reco_mode_options: Tuple[bool, ...] = (
                    # True,
                    False,
                )
                for use_tag_reco_mode in use_reco_mode_options:
                    if use_tag_reco_mode:
                        _sb_info_with_sys = SidebandInfo.create_modified_comparer_info(
                            comparer_info=sb_info_with_sys,
                            reco_mode_info_collection=BTagRecoModes.b_tag_by_groups_and_by_b_charge,
                            file_name_appendix="inFEITagModes",
                        )
                        data_mc_comparer = DataMCComparer(comparer_info=_sb_info_with_sys)
                    else:
                        data_mc_comparer = DataMCComparer(comparer_info=sb_info_with_sys)

                    data_mc_comparer.plot_data_mc_comparison(
                        mc_df=rescaled_mc_sample,
                        data_df=data_sample,
                        dry_run=False,
                        export_path=sb_plot_output_path,
                        weight_col=fitter.mc_weight_col_name,
                        reco_mode_col=TagRecoModeInfos.fei_reco_mode_col_name if use_tag_reco_mode else None,
                        style="stacked",
                        normalize_mc_to_data=False,
                        ratio_type="vs_uncert",
                        gof_check_method="toys",
                        adaptive_binning=False,
                        show_plots=False,
                    )

    def save_results(
        self,
        selection_name: str,
        fit_scenario_name: str,
        result: ContCalEvalResult,
    ) -> None:
        _file_name: str = f"{fit_scenario_name}__{selection_name}__{self.fit_obs_setup_tag}.json"
        output_tmp_path: PathType = os.path.join(self.output_path, f"tmp_{_file_name}")
        output_path: PathType = os.path.join(self.output_path, _file_name)

        result.save_to_file(file_path=output_tmp_path)
        os.rename(src=output_tmp_path, dst=output_path)

    def get_file_path_from_key(
        self,
        key: str,
        selection_name: str,
    ) -> PathType:
        return get_file_path_for(
            base_path=self.base_input_path,
            selection_name=selection_name,
            sample_key=key,
            sb_tag="_csmva_sideband",
            use_hammered_samples=True,
        )

    def load_mc(self, selection_name: str) -> pd.DataFrame:
        mc_paths = [self.get_file_path_from_key(key=k, selection_name=selection_name) for k in alternative_sample_keys]
        _df = pd.concat([pd.read_hdf(mcp) for mcp in mc_paths], ignore_index=True, sort=False).reset_index(drop=True)
        TagRecoModeInfos.add_new_fei_reco_mode_col(df=_df, check_if_new=True)

        n_e_ecl_extra_selection.apply_cut(df=_df)
        _df = _df.reset_index(drop=True)

        return apply_fei_calibration(df=_df)

    def load_data(self, selection_name: str) -> pd.DataFrame:
        data_path: PathType = self.get_file_path_from_key(key="DataY4S", selection_name=selection_name)
        _df = pd.read_hdf(data_path).reset_index(drop=True)
        TagRecoModeInfos.add_new_fei_reco_mode_col(df=_df, check_if_new=True)

        n_e_ecl_extra_selection.apply_cut(df=_df)

        return _df.reset_index(drop=True)


# endregion
