"""
Definition of fit routine class which executes the new dedicated fit.

Felix Metzner 2023
"""

import os
import copy

import numpy as np
import pandas as pd

from iminuit import Minuit
from iminuit.util import FMin
from dataclasses import dataclass
from collections.abc import MutableMapping
from typing import Tuple, List, Dict, Optional, Iterator, Generator

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis import systematics as rdstar_systematics_helpers
from rdstar.offline_analysis.observables.sig_id_observables import SigIDObservables
from rdstar.offline_analysis.rdstar_samples import alternative_sample_keys, main_sample_keys, y4s_data_key

from rdstar.offline_analysis.corrections.particle_id_corrections import ParticleIDCorrection
from rdstar.offline_analysis.corrections.qed_correction import get_coulomb_corr_factor, new_coulomb_correction_col_name

from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import get_file_path_for, get_file_base_name_for

# Calibrations
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_evaluation import apply_fei_calibration
from rdstar.offline_analysis.fitting.dedicated_fit_approach.dss_cal_evaluation import apply_dss_calibration
from rdstar.offline_analysis.fitting.dedicated_fit_approach.cont_cal_evaluation import apply_continuum_calibration

# Additional Selection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.tag_cal_fitters import right_sig_tag_combination_selection

# Fit
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fitter_rdstar import RDStarFitter
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import FitSetupCollection
from rdstar.offline_analysis.fitting.dedicated_fit_approach.rdstar_systematics import BKGHandling
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import FitSetupInfoContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_tuples_preparation import FitInfoForExternalFit
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import FitSetupTriplet
from rdstar.offline_analysis.fitting.dedicated_fit_approach.dedicated_sys_handler import (
    BKGHandlingInfoContainer,
    RDStarSystematicsContainer,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.plotting_tools import (
    NuisancePullPlotter,
    SystematicsShapePlotter,
    ShapePlotInfoContainer,
    SpecificShapePlotInfoContainer,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_result_container import RDStarFitResultContainer
from rdstar.offline_analysis.fitting.dedicated_fit_approach.rdstar_fit_eval.rdstar_money_plot import RDStarMoneyPlotter
from rdstar.offline_analysis.fitting.dedicated_fit_approach.rdstar_fit_eval.rdstar_fit_binning_plots import (
    RDStarFitBinningPlots,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.rdstar_fit_eval.rdstar_fit_component_plots import (
    RDStarFitComponentPlots,
)


__all__ = [
    "main_rd_rdstar_fit_triplet",
    "FitTypeInfo",
    "RDStarFitDataManager",
    "FitterManager",
    "FitMinimizerManager",
    "RDStarFitEvaluator",
]


logger = log_to_default(name=__name__)


main_rd_rdstar_fit_triplet: FitSetupTriplet = FitSetupCollection.rdstar_fit_setup


@dataclass(frozen=True)
class FitTypeInfo:
    fit_type: str
    toy_size: Optional[int] = None
    toy_seed: Optional[int] = None

    def __post_init__(self) -> None:
        if self.fit_type not in FitTypeInfo.get_fit_type_dict().keys():
            valid_types: str = ", ".join([f"{k} ({v})" for k, v in FitTypeInfo.get_fit_type_dict().items()])
            raise TypeError(f"Parameter 'fit_type' must be one of {valid_types}, but provided was {self.fit_type}")
        if self.fit_type == "T":
            if self.toy_size is None or self.toy_seed is None:
                raise RuntimeError("Seed and size must be provided for toy (T) fit_type!")

    @staticmethod
    def get_fit_type_dict() -> Dict[str, str]:
        return {
            "D": "Data",
            "A": "Asimov",
            "T": "Toy",
        }

    @property
    def recorded_data(self) -> bool:
        return self.fit_type == "D"

    @property
    def asimov_data(self) -> bool:
        return self.fit_type == "A"

    @property
    def toy_data(self) -> bool:
        return self.fit_type == "T"

    @property
    def checked_toy_seed(self) -> int:
        assert self.toy_data and self.toy_seed is not None
        return self.toy_seed

    @property
    def checked_toy_size(self) -> int:
        assert self.toy_data and self.toy_size is not None
        return self.toy_size

    @property
    def info_str(self) -> str:
        if self.recorded_data:
            return "Data"
        elif self.asimov_data:
            return "Asimov"
        elif self.toy_data:
            return f"Toys{self.toy_size}s{self.toy_seed}"
        else:
            raise RuntimeError(f"Unknown data setup:\n\t{self.__repr__()}...")


class RDStarFitDataManager:
    def __init__(
        self,
        fit_setup: FitSetupTriplet,
        base_input_path: PathType,
        mc_cache_version_tag: Optional[str],
        mc_cache_dir_path: Optional[PathType],
        use_alternative_gap_setup: bool = True,
    ) -> None:

        self._fit_setup: FitSetupTriplet = fit_setup

        if mc_cache_version_tag is not None:
            assert mc_cache_dir_path is not None, (mc_cache_version_tag, mc_cache_dir_path)
        self._mc_cache_version_tag: Optional[str] = mc_cache_version_tag
        self._mc_cache_dir_path: Optional[PathType] = mc_cache_dir_path

        self.base_input_path: PathType = base_input_path
        self._fit_weight_column_name: str = fit_setup.asimov.weight_info.col_name
        self.use_alternative_gap_setup: bool = use_alternative_gap_setup

        self._selection_name: str = "main_with_mva_selection_sec_sel_mva"
        self._sideband_tag: Optional[str] = None
        self._use_hammered_samples: bool = True

        self._mc_sample: Optional[pd.DataFrame] = None
        self._mc_sample_2015_belle: Optional[pd.DataFrame] = None

        self._data_sample: Optional[pd.DataFrame] = None

    def get_mc_sample(self, use_2015_belle: bool) -> pd.DataFrame:
        if use_2015_belle:
            return self.mc_sample_2015_belle
        else:
            return self.mc_sample

    @property
    def mc_sample(self) -> pd.DataFrame:
        if self._mc_sample is None:
            self._mc_sample = self.load_mc_sample()
        return self._mc_sample

    @property
    def mc_sample_2015_belle(self) -> pd.DataFrame:
        if self._mc_sample_2015_belle is None:
            _mc_sample_2015_belle: pd.DataFrame = RDStarFitter.get_mc_df_reweighted_to_old_belle_values(
                df=self.mc_sample,
                fit_weight_col_name=self._fit_weight_column_name,
            )
            self._mc_sample_2015_belle = _mc_sample_2015_belle
        return self._mc_sample_2015_belle

    @property
    def data_sample(self) -> pd.DataFrame:
        if self._data_sample is None:
            self._data_sample = self.load_data_sample()
        return self._data_sample

    @property
    def mc_sample_keys(self) -> Tuple[str, ...]:
        if self.use_alternative_gap_setup:
            return tuple(alternative_sample_keys)
        else:
            return tuple(main_sample_keys)

    def load_mc_sample(self) -> pd.DataFrame:

        if self._mc_cache_version_tag is None:
            logger.info("Loading MC samples without cache.")
            return self.load_and_prepare_mc_samples()
        else:
            logger.info("Loading MC samples with cache.")
            return self.load_cached_mc_sample()

    def load_data_sample(self) -> pd.DataFrame:
        data_path = self.get_file_path_for_key(sample_key=y4s_data_key)
        _data_sample: pd.DataFrame = pd.read_hdf(data_path).reset_index(drop=True)
        _selected_data_sample: pd.DataFrame = self.get_sample_with_preselection(df=_data_sample)
        return _selected_data_sample

    def load_and_prepare_mc_samples(self) -> pd.DataFrame:
        mc_paths: Tuple[PathType, ...] = tuple(
            [self.get_file_path_for_key(sample_key=sample_key) for sample_key in self.mc_sample_keys]
        )
        _mc_dfs: List[pd.DataFrame] = [pd.read_hdf(mc_path) for mc_path in mc_paths]
        _mc_sample: pd.DataFrame = pd.concat(_mc_dfs, ignore_index=True, sort=False).reset_index(drop=True)
        del _mc_dfs

        SigIDObservables.add_sig_id_observables_definitions(df=_mc_sample)
        _selected_mc_sample: pd.DataFrame = self.get_sample_with_preselection(df=_mc_sample)
        del _mc_sample

        _mc_sample_with_fei_cal: pd.DataFrame = apply_fei_calibration(df=_selected_mc_sample)
        del _selected_mc_sample
        _mc_sample_with_cont_cal: pd.DataFrame = apply_continuum_calibration(df=_mc_sample_with_fei_cal)
        del _mc_sample_with_fei_cal
        _mc_sample_with_dss_cal: pd.DataFrame = apply_dss_calibration(df=_mc_sample_with_cont_cal)
        del _mc_sample_with_cont_cal

        self._fit_setup.asimov.prepare_mc_df(df=_mc_sample_with_dss_cal)

        final_mc_sample: pd.DataFrame = self._final_mc_sample_preparations(
            df=_mc_sample_with_dss_cal,
            add_new_lepton_id_variations=False,
        )

        return final_mc_sample

    def load_cached_mc_sample(self) -> pd.DataFrame:
        assert self._mc_cache_dir_path is not None

        mc_cache_file_base_name: str = get_file_base_name_for(
            selection_name=self._selection_name,
            sb_tag=self._sideband_tag,
            use_hammered_samples=self._use_hammered_samples,
        )
        mc_cache_file_name: str = f"cached_mc_fit_sample_{mc_cache_file_base_name}_{self._mc_cache_version_tag}.h5"
        mc_cache_file_path: PathType = os.path.join(self._mc_cache_dir_path, mc_cache_file_name)

        if os.path.isfile(mc_cache_file_path):
            logger.info(f"\tUsing cached MC samples version {self._mc_cache_version_tag}:\n\t{mc_cache_file_path}")
            _df = pd.read_hdf(mc_cache_file_path)
            assert isinstance(_df, pd.DataFrame)
            return _df
        else:
            logger.info(f"\tCreating MC samples cache version {self._mc_cache_version_tag}:\n\t{mc_cache_file_path}")
            os.makedirs(self._mc_cache_dir_path, exist_ok=True)
            mc_sample: pd.DataFrame = self.load_and_prepare_mc_samples()
            tmp_mc_cache_file_path: PathType = os.path.join(self._mc_cache_dir_path, f"tmp_{mc_cache_file_name}")
            assert not os.path.exists(tmp_mc_cache_file_path), tmp_mc_cache_file_path
            mc_sample.to_hdf(path_or_buf=tmp_mc_cache_file_path, key="cached_mc_fitting_sample")
            os.rename(src=tmp_mc_cache_file_path, dst=mc_cache_file_path)
            return mc_sample

    @staticmethod
    def get_sample_with_preselection(df: pd.DataFrame) -> pd.DataFrame:
        return right_sig_tag_combination_selection.get_df_with_cut(df=df)

    @staticmethod
    def _final_mc_sample_preparations(
        df: pd.DataFrame,
        add_new_lepton_id_variations: bool = False,
    ) -> pd.DataFrame:
        logger.info("Adding final touches to the mc data frames:")
        _new_df: pd.DataFrame = copy.deepcopy(df)

        new_lepton_id_base_col_name: Optional[str] = None

        if add_new_lepton_id_variations:
            new_lepton_id_base_col_name = "new_lepton_pid"

            number_of_l_id_vars: int = 200
            logger.info(f"\tAdding {number_of_l_id_vars} new lepton ID efficiency correction variations.")
            pid_corr = ParticleIDCorrection(number_of_variations=number_of_l_id_vars)
            pid_corr.standalone_apply_lepton_pid_corrections(
                sample=_new_df,
                new_base_column_name=new_lepton_id_base_col_name,
                apply_for_fakes=True,
            )

            old_lepton_id_ratio_col_name: str = rdstar_systematics_helpers.get_lepton_id_ratio_col_name_for(
                base_name=None,
            )
            new_lepton_id_ratio_col_name: str = rdstar_systematics_helpers.get_lepton_id_ratio_col_name_for(
                base_name=new_lepton_id_base_col_name,
            )

            old_nominal_lid_weights: np.ndarray = _new_df[old_lepton_id_ratio_col_name].values
            new_nominal_lid_weights: np.ndarray = _new_df[new_lepton_id_ratio_col_name].values
            assert not np.any(np.isnan(new_nominal_lid_weights)), np.sum(np.isnan(new_nominal_lid_weights))
            assert np.allclose(old_nominal_lid_weights, new_nominal_lid_weights), (
                np.sum(~np.isclose(old_nominal_lid_weights, new_nominal_lid_weights)),
                np.max(old_nominal_lid_weights - new_nominal_lid_weights),
            )

        logger.info("\tSplitting lepton id correction weight columns for truth and fakes.")
        rdstar_systematics_helpers.split_lepton_id_cols_for(
            df=_new_df,
            base_lepton_id_name=new_lepton_id_base_col_name,
        )

        logger.info(f"\tAdding new QED correction weights in new column '{new_coulomb_correction_col_name}'.")
        get_coulomb_corr_factor(
            df=_new_df,
            col_name=new_coulomb_correction_col_name,
            only_leptonic_taus=False,
            d_mesons=None,
            on_reco_sig_side=False,
        )

        return _new_df

    def get_file_path_for_key(self, sample_key: str) -> PathType:
        return get_file_path_for(
            base_path=self.base_input_path,
            selection_name=self._selection_name,
            sample_key=sample_key,
            sb_tag=self._sideband_tag,
            use_hammered_samples=self._use_hammered_samples,
        )


class FitterManager:
    def __init__(
        self,
        fit_setup: FitSetupTriplet,
        fit_tag: str,
        data_manager: RDStarFitDataManager,
        file_dict_directory_path: PathType,
        use_belle_2015_as_nominal: bool = False,
    ) -> None:

        self.fit_setup: FitSetupTriplet = fit_setup
        self.fit_tag: str = fit_tag

        self.file_dict_directory_path: PathType = file_dict_directory_path

        self.use_belle_2015_as_nominal: bool = use_belle_2015_as_nominal

        self._data_manager: RDStarFitDataManager = data_manager

        self._fit_instances: Dict[str, RDStarFitter] = {}

    @staticmethod
    def get_instance_id_str(
        fit_type_info: FitTypeInfo,
        systematics_hash: str,
    ) -> str:
        return f"{fit_type_info.info_str}_{systematics_hash}"

    def get_fit_instance_for(
        self,
        fit_type_info: FitTypeInfo,
        systematics_container: RDStarSystematicsContainer,
    ) -> RDStarFitter:
        id_str: str = self.get_instance_id_str(
            fit_type_info=fit_type_info,
            systematics_hash=systematics_container.id_hash,
        )

        if id_str in self._fit_instances:
            return self._fit_instances[id_str]

        if fit_type_info.recorded_data:
            fit_instance: RDStarFitter = self.setup_data_fit()
        elif fit_type_info.asimov_data:
            fit_instance = self.setup_asimov_fit()
        elif fit_type_info.toy_data:
            fit_instance = self.setup_toy_fit(
                toy_size=fit_type_info.checked_toy_size,
                toy_seed=fit_type_info.checked_toy_seed,
            )
        else:
            raise RuntimeError(f"Unknown fit type info {fit_type_info.info_str}:\n\t{fit_type_info.__repr__()}")

        fit_instance.set_systematics_states(
            states_infos=systematics_container.get_states(),
        )

        assert id_str not in self._fit_instances.keys(), id_str
        self._fit_instances[id_str] = fit_instance
        return self._fit_instances[id_str]

    @property
    def mc_sample(self) -> pd.DataFrame:
        if self.use_belle_2015_as_nominal:
            return self._data_manager.mc_sample_2015_belle
        else:
            return self._data_manager.mc_sample

    @property
    def data_sample(self) -> pd.DataFrame:
        return self._data_manager.data_sample

    def setup_asimov_fit(self) -> RDStarFitter:
        fitter = RDStarFitter(
            fit_setup_info=self.fit_setup.asimov,
            mc_df=self.mc_sample,
            data_df=None,
            file_dict_dir_path=self.file_dict_directory_path,
            use_old_ratio_values=self.use_belle_2015_as_nominal,
        )

        return fitter

    def setup_data_fit(self) -> RDStarFitter:
        fitter = RDStarFitter(
            fit_setup_info=self.fit_setup.data,
            mc_df=self.mc_sample,
            data_df=self.data_sample,
            file_dict_dir_path=self.file_dict_directory_path,
            use_old_ratio_values=self.use_belle_2015_as_nominal,
        )

        return fitter

    def setup_toy_fit(
        self,
        toy_size: int,
        toy_seed: int,
    ) -> RDStarFitter:
        toy_fitter_info: FitSetupInfoContainer = self.fit_setup.get_toy_setup(
            toy_size=toy_size,
            toy_seed=toy_seed,
        )

        fitter = RDStarFitter(
            fit_setup_info=toy_fitter_info,
            mc_df=self.mc_sample,
            data_df=None,
            file_dict_dir_path=self.file_dict_directory_path,
            use_old_ratio_values=self.use_belle_2015_as_nominal,
        )

        return fitter

    def produce_component_plots(
        self,
        output_dir_path: PathType,
    ) -> None:
        component_plotter = RDStarFitComponentPlots(
            evaluation_name=self.fit_tag,
            base_output_path=output_dir_path,
            using_old_ratio_values=self.use_belle_2015_as_nominal,
        )
        component_plotter.produce_fit_component_plots(
            mc_sample=self.mc_sample,
            close_figures=False,
        )

    def produce_binning_plots(
        self,
        output_dir_path: PathType,
        for_data: bool = False,
    ) -> None:
        binning_plotter = RDStarFitBinningPlots(
            evaluation_name=self.fit_tag,
            base_output_path=output_dir_path,
        )
        binning_plotter.produce_fit_binning_plots(
            df=self.mc_sample,
            is_mc=True,
        )

        if for_data:
            binning_plotter.produce_fit_binning_plots(
                df=self._data_manager.data_sample,
                is_mc=False,
            )


class FitMinimizerManager:
    def __init__(
        self,
        fit_instance: RDStarFitter,
    ) -> None:
        self.fit_instance: RDStarFitter = fit_instance

        self._initial_values: Optional[np.ndarray] = None
        self._minimizer: Optional[Minuit] = None

    @property
    def initial_values(self) -> np.ndarray:
        if self._initial_values is None:
            self.initial_values = np.array(list(self.fit_instance.initial_parameters))
        return self._initial_values

    @initial_values.setter
    def initial_values(self, parameter_values: np.ndarray) -> None:
        self._initial_values = parameter_values

    @property
    def minimizer(self) -> Minuit:
        if self._minimizer is None:
            self._minimizer = self.get_minimizer()
        return self._minimizer

    def get_minimizer(self) -> Minuit:
        minuit_instance = Minuit(
            self.fit_instance.cost_fcn,
            self.initial_values,
            name=tuple([str(i) for i in range(len(list(self.initial_values)))]),
        )

        self.fit_instance.set_iminuit_properties(minuit_instance=minuit_instance)

        return minuit_instance

    def run_fit(
        self,
        n_calls: Optional[int] = None,
        n_iterate: Optional[int] = None,
        run_minos: bool = True,
    ) -> FMin:
        m: Minuit = self.minimizer

        if run_minos:
            f_min = (
                m.simplex()
                .migrad(
                    ncall=self.fit_instance.minuit_settings.n_call if n_calls is None else n_calls,
                    iterate=self.fit_instance.minuit_settings.n_iterate if n_iterate is None else n_iterate,
                )
                .minos()
                .hesse()
                .fmin
            )
        else:
            f_min = (
                m.simplex()
                .migrad(
                    ncall=self.fit_instance.minuit_settings.n_call if n_calls is None else n_calls,
                    iterate=self.fit_instance.minuit_settings.n_iterate if n_iterate is None else n_iterate,
                )
                .hesse()
                .fmin
            )

        return f_min


@dataclass(frozen=True)
class RDStarFitStatusContainer:
    name: str
    fit_type_info: FitTypeInfo
    bkg_handling_info_str: str
    fitter_instance: RDStarFitter
    minuit_instance: Minuit
    fit_result: RDStarFitResultContainer


@dataclass(frozen=True)
class SysResultItem:
    name: str
    sys_names: Tuple[str, ...]
    rd_error: float
    rd_relative_error: float
    rd_star_error: float
    rd_star_relative_error: float

    @staticmethod
    def create_from(
        name: str,
        sys_names: Tuple[str, ...],
        this_result: RDStarFitResultContainer,
        base_result: RDStarFitResultContainer,
        ignore_check: bool = False,
    ) -> "SysResultItem":

        if ignore_check:
            if not np.all(np.isclose(this_result.param_values[:2], base_result.param_values[:2], rtol=1e-5, atol=1.0e-4)):
                logger.warning(
                    f"Values for ignored check:\n"
                    f"\t This result: {tuple(this_result.param_values[:2])},\n"
                    f"\t Base result: {tuple(base_result.param_values[:2])}"
                )
        else:
            assert np.all(
                np.isclose(this_result.param_values[:2], base_result.param_values[:2], rtol=1e-5, atol=1.0e-4)
            ), (
                tuple(this_result.param_values[:2]),
                tuple(base_result.param_values[:2]),
            )

        errors: np.ndarray = np.sqrt(np.abs(np.square(base_result.errors[:2]) - np.square(this_result.errors[:2])))
        relative_errors: np.ndarray = errors / base_result.param_values[:2]

        rd_error: float = errors[0]
        rd_relative_error: float = relative_errors[0]
        rd_star_error: float = errors[1]
        rd_star_relative_error: float = relative_errors[1]

        return SysResultItem(
            name=name,
            sys_names=sys_names,
            rd_error=rd_error,
            rd_relative_error=rd_relative_error,
            rd_star_error=rd_star_error,
            rd_star_relative_error=rd_star_relative_error,
        )


class SysResultContainer(MutableMapping):
    def __init__(self, name: str) -> None:
        self._name: str = name
        self._sy_results: Dict[str, SysResultItem] = dict()

    @property
    def name(self) -> str:
        return self._name

    def __getitem__(self, item: str) -> SysResultItem:
        if item not in self._sy_results:
            raise KeyError(item)
        return self._sy_results[item]

    def __setitem__(self, key: str, value: SysResultItem) -> None:
        if key in self._sy_results:
            raise KeyError(f"key {key} already exists!")
        if not isinstance(value, SysResultItem):
            raise TypeError(f"value is not of type 'SysResultItem' but of type '{type(value)}'")
        self._sy_results.update({key: value})

    def __delitem__(self, item: str) -> None:
        if item not in self._sy_results:
            raise KeyError(item)
        raise RuntimeError("SysResultContainer is append-only.")

    def __iter__(self) -> Iterator[str]:
        return iter(self._sy_results)

    def __len__(self) -> int:
        return len(self._sy_results)

    @staticmethod
    def get_name_from(sys_names: Tuple[str, ...]) -> str:
        return "_".join(sys_names)

    def add_sys_result(
        self,
        sys_names: Tuple[str, ...],
        this_result: RDStarFitResultContainer,
        base_result: RDStarFitResultContainer,
        ignore_check: bool = False,
    ) -> str:
        _name: str = self.get_name_from(sys_names=sys_names)
        sys_res_item: SysResultItem = SysResultItem.create_from(
            name=_name,
            sys_names=sys_names,
            this_result=this_result,
            base_result=base_result,
            ignore_check=ignore_check,
        )
        self[sys_res_item.name] = sys_res_item
        return sys_res_item.name


class RDStarFitEvaluator:
    def __init__(
        self,
        fit_setup: FitSetupTriplet,
        fit_tag: str,
        base_output_dir_path: PathType,
        base_input_path: PathType,
        mc_cache_version_tag: Optional[str],
        mc_cache_dir_path: Optional[PathType],
        file_dict_directory_path: PathType,
        use_alternative_gap_setup: bool = True,
        run_minos: bool = True,
    ) -> None:
        self.fit_setup: FitSetupTriplet = fit_setup
        self.fit_tag: str = fit_tag

        self.base_output_dir_path: PathType = base_output_dir_path

        self.base_input_path: PathType = base_input_path
        self.file_dict_directory_path: PathType = file_dict_directory_path

        self.run_minos: bool = run_minos

        self._data_manager = RDStarFitDataManager(
            fit_setup=fit_setup,
            base_input_path=self.base_input_path,
            mc_cache_version_tag=mc_cache_version_tag,
            mc_cache_dir_path=mc_cache_dir_path,
            use_alternative_gap_setup=use_alternative_gap_setup,
        )

        self._fit_manager_b23: FitterManager = FitterManager(
            fit_setup=self.fit_setup,
            fit_tag=self.fit_tag,
            data_manager=self._data_manager,
            file_dict_directory_path=self.file_dict_directory_path,
            use_belle_2015_as_nominal=False,
        )

        self._fit_manager_b15: FitterManager = FitterManager(
            fit_setup=self.fit_setup,
            fit_tag=self.fit_tag,
            data_manager=self._data_manager,
            file_dict_directory_path=self.file_dict_directory_path,
            use_belle_2015_as_nominal=True,
        )

        self._full_systematics_container: RDStarSystematicsContainer = RDStarFitter.full_systematics_container
        self._empty_systematics_container: RDStarSystematicsContainer = RDStarFitter.empty_systematics_container

        self._fit_container: Dict[str, RDStarFitStatusContainer] = {}
        self._fit_sys_containers: Dict[str, SysResultContainer] = {}
        self._fit_bkg_sys_containers: Dict[str, SysResultContainer] = {}

    @property
    def full_systematics_container(self) -> RDStarSystematicsContainer:
        return self._full_systematics_container

    @property
    def empty_systematics_container(self) -> RDStarSystematicsContainer:
        return self._empty_systematics_container

    def get_fit_manager_with(self, belle15_as_nominal: bool) -> FitterManager:
        if belle15_as_nominal:
            return self._fit_manager_b15
        else:
            return self._fit_manager_b23

    @staticmethod
    def get_fit_name(
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        bkg_handling_infos: BKGHandlingInfoContainer,
        systematics_container: RDStarSystematicsContainer,
    ) -> str:
        t_info: str = fit_type_info.info_str
        b15_info: str = "b15" if use_b2015_as_nominal else "b24"

        systematics_names: Tuple[str, ...] = systematics_container.systematics_names
        stat_sys_info: str = "with_sys" if systematics_names else "stat_only"

        bkg_handling_info_str: str = bkg_handling_infos.values_as_info_str

        sys_bools: str = systematics_container.id_hash
        return f"{base_name}_{t_info}_{b15_info}_{bkg_handling_info_str}_{stat_sys_info}_syshash{sys_bools}"

    def run_fit(
        self,
        rerun_fit: bool,
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        bkg_handling_infos: BKGHandlingInfoContainer,
        systematics_container: RDStarSystematicsContainer,
    ) -> RDStarFitStatusContainer:
        fit_name: str = self.get_fit_name(
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=use_b2015_as_nominal,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=systematics_container,
        )

        if rerun_fit or fit_name not in self._fit_container:

            fit_manager = self.get_fit_manager_with(belle15_as_nominal=use_b2015_as_nominal)
            fit_instance: RDStarFitter = fit_manager.get_fit_instance_for(
                fit_type_info=fit_type_info,
                systematics_container=systematics_container,
            )

            minimizer_manager = FitMinimizerManager(fit_instance=fit_instance)

            for sys_name, sys_info in systematics_container.items():
                if not sys_info.is_active:
                    for sys_param_id in sys_info.nuisance_param_ids:
                        minimizer_manager.minimizer.fixed[sys_param_id] = True
                else:
                    for sys_param_id in sys_info.nuisance_param_ids:
                        minimizer_manager.minimizer.fixed[sys_param_id] = False

            for always_fixed_param_id in systematics_container.always_fixed_nuisance_parameter_indices:
                minimizer_manager.minimizer.fixed[always_fixed_param_id] = True

            for bkg_info in bkg_handling_infos.values():
                for bkg_yield_param_id, bkg_constraint_param_id_offset in zip(
                    bkg_info.bkg_yield_ids,
                    bkg_info.bkg_constraint_id_offsets,
                ):
                    bkg_constraint_param_id: int = (
                        bkg_handling_infos.bkg_constraint_param_id_start + bkg_constraint_param_id_offset
                    )
                    if bkg_info.status == BKGHandling.BKG_FIXED:
                        minimizer_manager.minimizer.fixed[bkg_yield_param_id] = True
                        minimizer_manager.minimizer.fixed[bkg_constraint_param_id] = True
                    elif bkg_info.status == BKGHandling.BKG_FLOATING:
                        minimizer_manager.minimizer.fixed[bkg_yield_param_id] = False
                        minimizer_manager.minimizer.fixed[bkg_constraint_param_id] = True
                    elif bkg_info.status == BKGHandling.BKG_CONSTRAINT:
                        minimizer_manager.minimizer.fixed[bkg_yield_param_id] = True
                        minimizer_manager.minimizer.fixed[bkg_constraint_param_id] = False
                    else:
                        raise ValueError(
                            f"The bkg_handling case '{bkg_info.status.name}: {bkg_info.status.value}' is not recognized"
                        )

            minimizer_manager.run_fit(
                n_calls=fit_instance.minuit_settings.n_call,
                n_iterate=fit_instance.minuit_settings.n_iterate,
                run_minos=self.run_minos,
            )

            fit_result: RDStarFitResultContainer = RDStarFitResultContainer.from_minimizer_and_fitter(
                m=minimizer_manager.minimizer,
                f=fit_instance,
            )

            fit_status = RDStarFitStatusContainer(
                name=fit_name,
                fit_type_info=fit_type_info,
                bkg_handling_info_str=bkg_handling_infos.values_as_info_str,
                fitter_instance=fit_instance,
                minuit_instance=minimizer_manager.minimizer,
                fit_result=fit_result,
            )

            self._fit_container[fit_name] = fit_status

        return self._fit_container[fit_name]

    def get_fit(
        self,
        run_if_necessary: bool,
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        bkg_handling_infos: BKGHandlingInfoContainer,
        systematics_container: RDStarSystematicsContainer,
    ) -> RDStarFitStatusContainer:
        fit_name: str = self.get_fit_name(
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=use_b2015_as_nominal,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=systematics_container,
        )

        if fit_name not in self._fit_container:
            if run_if_necessary:
                return self.run_fit(
                    rerun_fit=False,
                    base_name=base_name,
                    fit_type_info=fit_type_info,
                    use_b2015_as_nominal=use_b2015_as_nominal,
                    bkg_handling_infos=bkg_handling_infos,
                    systematics_container=systematics_container,
                )
            else:
                raise KeyError(f"Result for fit {fit_name} not available, yet! Try calling 'run_fit'!")

        return self._fit_container[fit_name]

    def produce_money_plot(
        self,
        base_name: str,
        fit_type_info: FitTypeInfo,
        bkg_handling_infos: BKGHandlingInfoContainer,
        sm_label: Optional[str] = None,
        np_label: Optional[str] = None,
    ) -> None:
        full_rd_rdstar_sm_status: RDStarFitStatusContainer = self.get_fit(
            run_if_necessary=True,
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=False,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=self.full_systematics_container,
        )
        stat_rd_rdstar_sm_status: RDStarFitStatusContainer = self.get_fit(
            run_if_necessary=True,
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=False,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=self.empty_systematics_container,  # Stat. uncertainty only! -> No systematics!
        )

        full_rd_rdstar_np_status: RDStarFitStatusContainer = self.get_fit(
            run_if_necessary=True,
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=True,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=self.full_systematics_container,
        )
        stat_rd_rdstar_np_status: RDStarFitStatusContainer = self.get_fit(
            run_if_necessary=True,
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=True,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=self.empty_systematics_container,  # Stat. uncertainty only! -> No systematics!
        )

        bkg_handling_info_str: str = bkg_handling_infos.values_as_info_str

        money_plotter = RDStarMoneyPlotter(
            evaluation_name=f"{self.fit_tag}_{fit_type_info.info_str}_{bkg_handling_info_str}",
            base_output_path=self.base_output_dir_path,
            full_rd_rdstar_sm_result=full_rd_rdstar_sm_status.fit_result,
            stat_rd_rdstar_sm_result=stat_rd_rdstar_sm_status.fit_result,
            full_rd_rdstar_np_result=full_rd_rdstar_np_status.fit_result,
            stat_rd_rdstar_np_result=stat_rd_rdstar_np_status.fit_result,
            this_sm_label=sm_label,
            this_np_label=np_label,
        )

        money_plotter.produce_significance_plot()

    def evaluate_systematics_for(
        self,
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        bkg_handling_infos: BKGHandlingInfoContainer,
        ignore_check: bool = False,
    ) -> SysResultContainer:
        b15_info: str = "b15" if use_b2015_as_nominal else "b24"
        fit_sys_container_key: str = f"{base_name}_{b15_info}_fit_sys_result_container"

        if fit_sys_container_key in self._fit_sys_containers:
            _fit_sys_container: SysResultContainer = self._fit_sys_containers[fit_sys_container_key]
        else:
            _fit_sys_container = self._evaluate_fit_sys_effects(
                container_name=fit_sys_container_key,
                base_name=base_name,
                fit_type_info=fit_type_info,
                use_b2015_as_nominal=use_b2015_as_nominal,
                bkg_handling_infos=bkg_handling_infos,
                ignore_check=ignore_check,
            )

        if fit_sys_container_key not in self._fit_sys_containers:
            self._fit_sys_containers[fit_sys_container_key] = _fit_sys_container

        return _fit_sys_container

    def _evaluate_fit_sys_effects(
        self,
        container_name: str,
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        bkg_handling_infos: BKGHandlingInfoContainer,
        ignore_check: bool = False,
    ) -> SysResultContainer:
        _fit_sys_container = SysResultContainer(name=container_name)

        base_fit_status: RDStarFitStatusContainer = self.get_fit(
            run_if_necessary=True,
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=use_b2015_as_nominal,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=self.full_systematics_container,
        )

        logger.info("Evaluating systematic effects for")

        for sys_name, sys_container in self.full_systematics_container.get_sys_setup_infos_for_evaluation():
            assert isinstance(sys_name, str), (sys_name, type(sys_name))
            assert isinstance(sys_container, RDStarSystematicsContainer), (sys_name, type(sys_container))

            sys_names: Tuple[str, ...] = (sys_name,)
            logger.info(f"\tsys_name = {sys_name}")

            this_fit_status: RDStarFitStatusContainer = self.get_fit(
                run_if_necessary=True,
                base_name=base_name,
                fit_type_info=fit_type_info,
                use_b2015_as_nominal=use_b2015_as_nominal,
                bkg_handling_infos=bkg_handling_infos,
                systematics_container=sys_container,
            )

            _fit_sys_container.add_sys_result(
                sys_names=sys_names,
                this_result=this_fit_status.fit_result,
                base_result=base_fit_status.fit_result,
                ignore_check=ignore_check,
            )

            _r: SysResultItem = _fit_sys_container[_fit_sys_container.get_name_from(sys_names=sys_names)]
            logger.info(
                f"\tResults for sys_name = {sys_name}\n"
                f"\t\t the results are\n"
                f"\t\t\t sigma R(D) = {_r.rd_error} (absolute) or {_r.rd_relative_error * 100} % (relative)\n"
                f"\t\t\t sigma R(D*) = {_r.rd_star_error} (absolute) or {_r.rd_star_relative_error * 100} % (relative)"
            )

        return _fit_sys_container

    def evaluate_bkg_setup_systematics_for(
        self,
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        systematics_container: RDStarSystematicsContainer,
        ignore_check: bool = False,
    ) -> SysResultContainer:
        b15_info: str = "b15" if use_b2015_as_nominal else "b24"
        fit_bkg_sys_container_key: str = f"{base_name}_{b15_info}_fit_bkg_sys_result_container"

        if fit_bkg_sys_container_key in self._fit_bkg_sys_containers:
            _fit_bkg_sys_container: SysResultContainer = self._fit_bkg_sys_containers[fit_bkg_sys_container_key]
        else:
            _fit_bkg_sys_container = self._evaluate_fit_bkg_constraint_sys_effects(
                container_name=fit_bkg_sys_container_key,
                base_name=base_name,
                fit_type_info=fit_type_info,
                use_b2015_as_nominal=use_b2015_as_nominal,
                systematics_container=systematics_container,
                ignore_check=ignore_check,
            )

        if fit_bkg_sys_container_key not in self._fit_bkg_sys_containers:
            self._fit_bkg_sys_containers[fit_bkg_sys_container_key] = _fit_bkg_sys_container

        return _fit_bkg_sys_container

    def _evaluate_fit_bkg_constraint_sys_effects(
        self,
        container_name: str,
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        systematics_container: RDStarSystematicsContainer,
        ignore_check: bool = False,
    ) -> SysResultContainer:
        _fit_bkg_sys_container = SysResultContainer(name=container_name)

        base_fit_status: RDStarFitStatusContainer = self.get_fit(
            run_if_necessary=True,
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=use_b2015_as_nominal,
            bkg_handling_infos=RDStarFitter.basic_bkg_handler,
            systematics_container=systematics_container,
        )

        logger.info("Evaluating background constraint systematic effects for")

        for bkg_name, bkg_setup_info in RDStarFitter.basic_bkg_handler.get_bkg_handler_infos_for_evaluation():
            assert isinstance(bkg_name, str), (bkg_name, type(bkg_name))
            assert isinstance(bkg_setup_info, BKGHandlingInfoContainer), (bkg_name, type(bkg_setup_info))

            bkg_names: Tuple[str, ...] = (bkg_name,)
            logger.info(f"\tbkg_name = {bkg_name}")

            this_fit_status: RDStarFitStatusContainer = self.get_fit(
                run_if_necessary=True,
                base_name=base_name,
                fit_type_info=fit_type_info,
                use_b2015_as_nominal=use_b2015_as_nominal,
                bkg_handling_infos=bkg_setup_info,
                systematics_container=systematics_container,
            )

            _fit_bkg_sys_container.add_sys_result(
                sys_names=bkg_names,
                this_result=this_fit_status.fit_result,
                base_result=base_fit_status.fit_result,
                ignore_check=ignore_check,
            )

            _r: SysResultItem = _fit_bkg_sys_container[_fit_bkg_sys_container.get_name_from(sys_names=bkg_names)]
            logger.info(
                f"\tResults for bkg_name = {bkg_name}\n"
                f"\t\t the results are\n"
                f"\t\t\t sigma R(D) = {_r.rd_error} (absolute) or {_r.rd_relative_error * 100} % (relative)\n"
                f"\t\t\t sigma R(D*) = {_r.rd_star_error} (absolute) or {_r.rd_star_relative_error * 100} % (relative)"
            )

        return _fit_bkg_sys_container

    def plot_nuisance_pulls(
        self,
        fit_result: RDStarFitResultContainer,
    ) -> None:
        nuisance_pull_plotter: NuisancePullPlotter = NuisancePullPlotter(base_output_dir_path=self.base_output_dir_path)

        logger.info("Creating nuisance pull overview plot for additive systematics.")
        sys_infos = tuple(v for v in self.full_systematics_container.values())
        nuisance_pull_plotter.create_nuisance_pull_overview_plots(
            additive=True,
            sys_infos=sys_infos,
            fit_result=fit_result,
        )

        logger.info("Creating nuisance pull overview plot for multiplicative systematics.")
        nuisance_pull_plotter.create_nuisance_pull_overview_plots(
            additive=False,
            sys_infos=sys_infos,
            fit_result=fit_result,
        )

        logger.info(f"Creating single nuisance pull plots and saving them to\n\t{nuisance_pull_plotter.output_dir_path}")

        for sys_name, sys_container in self.full_systematics_container.items():
            if sys_name == RDStarFitter.full_systematics_container.add_sys__mc_statistics.systematics_key:
                logger.info(f"\tSkipping nuisance pull plot for {sys_name}")
                continue
            logger.info(f"\tPlotting nuisance pulls for {sys_name}")
            nuisance_pull_plotter.create_nuisance_pull_plot_from(
                sys_info=sys_container,
                fit_result=fit_result,
            )

    def plot_sig_ff_sys_shape_effects(
        self,
        fit_status_container: RDStarFitStatusContainer,
        add_statistical_uncertainties: bool,
        normalize_pulls: bool,
        plot_subsets: bool,
        scale_factor: Optional[float] = None,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:
        sys_shape_plotter: SystematicsShapePlotter = SystematicsShapePlotter(
            base_output_dir_path=self.base_output_dir_path,
            fit_setup=self.fit_setup,
        )

        fit_instance: RDStarFitter = fit_status_container.fitter_instance

        example_sys_shape_info = ShapePlotInfoContainer(
            name=RDStarFitter.full_systematics_container.add_sys__ff_signal.systematics_key,
            latex_str=RDStarFitter.full_systematics_container.add_sys__ff_signal.latex_str,
            normed_base_shape=fit_instance.template_shapes,
            relative_shape_error=fit_instance.ff_sig_rel_shape_sys_uncert_matrix,
            pure_bin_counts=fit_instance.squared_template_stat_error,
            components_to_consider=("BpDztau", "BzDmtau", "BpDzStau", "BzDmStau"),
            scale_factor=scale_factor,
            plot_subsets=plot_subsets,
        )

        yield from sys_shape_plotter.create_systematics_shape_plots(
            sys_shape_info=example_sys_shape_info,
            # add_statistical_uncertainties=add_statistical_uncertainties,
            # normalize_pulls=normalize_pulls,
        )

    def plot_norm_ff_sys_shape_effects(
        self,
        fit_status_container: RDStarFitStatusContainer,
        add_statistical_uncertainties: bool,
        normalize_pulls: bool,
        plot_subsets: bool,
        scale_factor: Optional[float] = None,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:
        sys_shape_plotter: SystematicsShapePlotter = SystematicsShapePlotter(
            base_output_dir_path=self.base_output_dir_path,
            fit_setup=self.fit_setup,
        )

        fit_instance: RDStarFitter = fit_status_container.fitter_instance

        example_sys_shape_info = ShapePlotInfoContainer(
            name=RDStarFitter.full_systematics_container.add_sys__ff_norm.systematics_key,
            latex_str=RDStarFitter.full_systematics_container.add_sys__ff_norm.latex_str,
            normed_base_shape=fit_instance.template_shapes,
            relative_shape_error=fit_instance.ff_norm_rel_shape_sys_uncert_matrix,
            pure_bin_counts=fit_instance.squared_template_stat_error,
            components_to_consider=("BpDzl", "BzDml", "BpDzSl", "BzDmSl"),
            scale_factor=scale_factor,
            plot_subsets=plot_subsets,
        )

        yield from sys_shape_plotter.create_systematics_shape_plots(
            sys_shape_info=example_sys_shape_info,
            # add_statistical_uncertainties=add_statistical_uncertainties,
            # normalize_pulls=normalize_pulls,
        )

    def plot_l_id_sys_shape_effects(
        self,
        fit_status_container: RDStarFitStatusContainer,
        add_statistical_uncertainties: bool,
        normalize_pulls: bool,
        plot_subsets: bool,
        scale_factor: Optional[float] = None,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:
        sys_shape_plotter: SystematicsShapePlotter = SystematicsShapePlotter(
            base_output_dir_path=self.base_output_dir_path,
            fit_setup=self.fit_setup,
        )

        fit_instance: RDStarFitter = fit_status_container.fitter_instance

        example_sys_shape_info = ShapePlotInfoContainer(
            name=RDStarFitter.full_systematics_container.add_sys__pid_lepton.systematics_key,
            latex_str=RDStarFitter.full_systematics_container.add_sys__pid_lepton.latex_str,
            normed_base_shape=fit_instance.template_shapes,
            relative_shape_error=fit_instance.lepton_id_rel_shape_sys_uncert_matrix,
            pure_bin_counts=fit_instance.squared_template_stat_error,
            components_to_consider=("BpDzl", "BzDml", "BpDzSl", "BzDmSl", "BBbarBKG_in_cB", "BBbarBKG_in_nB"),
            scale_factor=scale_factor,
            plot_subsets=plot_subsets,
        )

        yield from sys_shape_plotter.create_systematics_shape_plots(
            sys_shape_info=example_sys_shape_info,
            # add_statistical_uncertainties=add_statistical_uncertainties,
            # normalize_pulls=normalize_pulls,
        )

    def plot_tracking_sys_shape_effects(
        self,
        fit_status_container: RDStarFitStatusContainer,
        add_statistical_uncertainties: bool,
        normalize_pulls: bool,
        plot_subsets: bool,
        scale_factor: Optional[float] = None,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:
        sys_shape_plotter: SystematicsShapePlotter = SystematicsShapePlotter(
            base_output_dir_path=self.base_output_dir_path,
            fit_setup=self.fit_setup,
        )

        fit_instance: RDStarFitter = fit_status_container.fitter_instance

        example_sys_shape_info = ShapePlotInfoContainer(
            name=RDStarFitter.full_systematics_container.add_sys__tracking.systematics_key,
            latex_str=RDStarFitter.full_systematics_container.add_sys__tracking.latex_str,
            normed_base_shape=fit_instance.template_shapes,
            relative_shape_error=fit_instance.tracking_rel_shape_sys_effect,
            pure_bin_counts=fit_instance.squared_template_stat_error,
            components_to_consider=None,
            scale_factor=scale_factor,
            plot_subsets=plot_subsets,
        )

        yield from sys_shape_plotter.create_systematics_shape_plots(
            sys_shape_info=example_sys_shape_info,
            # add_statistical_uncertainties=add_statistical_uncertainties,
            # normalize_pulls=normalize_pulls,
        )

    def plot_k_short_sys_shape_effects(
        self,
        fit_status_container: RDStarFitStatusContainer,
        add_statistical_uncertainties: bool,
        normalize_pulls: bool,
        plot_subsets: bool,
        scale_factor: Optional[float] = None,
    ) -> Generator[SpecificShapePlotInfoContainer, None, None]:
        sys_shape_plotter: SystematicsShapePlotter = SystematicsShapePlotter(
            base_output_dir_path=self.base_output_dir_path,
            fit_setup=self.fit_setup,
        )

        fit_instance: RDStarFitter = fit_status_container.fitter_instance

        example_sys_shape_info = ShapePlotInfoContainer(
            name=RDStarFitter.full_systematics_container.add_sys__k_short_eff.systematics_key,
            latex_str=RDStarFitter.full_systematics_container.add_sys__k_short_eff.latex_str,
            normed_base_shape=fit_instance.template_shapes,
            relative_shape_error=fit_instance.k_short_rel_shape_sys_uncert_matrix,
            pure_bin_counts=fit_instance.squared_template_stat_error,
            components_to_consider=None,
            scale_factor=scale_factor,
            plot_subsets=plot_subsets,
        )

        yield from sys_shape_plotter.create_systematics_shape_plots(
            sys_shape_info=example_sys_shape_info,
            # add_statistical_uncertainties=add_statistical_uncertainties,
            # normalize_pulls=normalize_pulls,
        )

    def dump_data_for_external_fitter(
        self,
        output_tag: str,
        output_dir_path: PathType,
        #
        base_name: str,
        fit_type_info: FitTypeInfo,
        use_b2015_as_nominal: bool,
        bkg_handling_infos: BKGHandlingInfoContainer,
        systematics_container: RDStarSystematicsContainer,
    ) -> None:
        fit_name: str = self.get_fit_name(
            base_name=base_name,
            fit_type_info=fit_type_info,
            use_b2015_as_nominal=use_b2015_as_nominal,
            bkg_handling_infos=bkg_handling_infos,
            systematics_container=systematics_container,
        )

        fit_data_file_name = f"fit_dataframe_{fit_name}_{output_tag}.h5"
        fit_data_file_path = os.path.join(output_dir_path, fit_data_file_name)
        tmp_fit_data_file_path = os.path.join(output_dir_path, f"tmp_{fit_data_file_name}")
        if os.path.exists(fit_data_file_path):
            raise FileExistsError(fit_data_file_path)
        assert not os.path.exists(tmp_fit_data_file_path), tmp_fit_data_file_path

        _fit_info: RDStarFitStatusContainer = self._fit_container[fit_name]
        rdstar_fit_instance: RDStarFitter = _fit_info.fitter_instance

        if fit_name not in self._fit_container:
            raise KeyError(f"Fit '{fit_name}' is not available, yet! Try running it first!")

        os.makedirs(output_dir_path, exist_ok=True)
        fit_info_file_path: PathType = os.path.join(output_dir_path, f"fit_info_{fit_name}_{output_tag}.pkl")

        _bin_edges_by_reco_ch_ids: Dict[int, Tuple[Tuple[float, ...], ...]] = {
            rch.ch_index: tuple(obs_infos.bin_edges for obs_infos in rch.observable_infos)
            for rch in rdstar_fit_instance.fit_setup_info.reco_channels
        }

        fit_info_container = FitInfoForExternalFit(
            base_tuples=rdstar_fit_instance.base_mc_bin_counts,
            channel_concatenated_tuples=rdstar_fit_instance.pure_mc_templates,
            normed_tuples=rdstar_fit_instance.template_shapes,
            #
            sig_ids_by_ordered_component_ids={
                c.comp_index: c.sig_ids for c in rdstar_fit_instance.fit_setup_info.components
            },
            ordered_reco_ch_infos={
                rch.ch_index: rch.reco_channel_ids for rch in rdstar_fit_instance.fit_setup_info.reco_channels
            },
            ordered_observable_cols=tuple(
                o.hist_var.df_label for o in rdstar_fit_instance.fit_setup_info.observable_infos
            ),
            #
            bin_edges_by_reco_ch_ids=_bin_edges_by_reco_ch_ids,
        )

        fit_info_container.save(file_path=str(fit_info_file_path))

        fit_data_frame: pd.DataFrame = rdstar_fit_instance.internal_mc_sample_df
        fit_data_frame.to_hdf(path_or_buf=tmp_fit_data_file_path, key="fitting")
        os.rename(src=tmp_fit_data_file_path, dst=fit_data_file_path)


# TODO: BF variation code snippet
__new_code_for_bf_sys: str = """
def __d_meson_decay_bf_uncertainties_example_code():
    df_mc["sig_d_decay_abs"] = df_mc["sig_d_decay"].apply(lambda x: tuple([abs(x_) for x_ in x]) if isinstance(x, tuple) else x)

    df_mc["D_BF_variation_nominal"] = 1
    df_mc["D_BF_variation_0_up"  ] = 1
    df_mc["D_BF_variation_0_down"] = 1
    df_mc["D_BF_variation_1_up"  ] = 1
    df_mc["D_BF_variation_1_down"] = 1
    df_mc["D_BF_variation_2_up"  ] = 1
    df_mc["D_BF_variation_2_down"] = 1
    df_mc["D_BF_variation_3_up"  ] = 1
    df_mc["D_BF_variation_3_down"] = 1
    df_mc["D_BF_variation_4_up"  ] = 1
    df_mc["D_BF_variation_4_down"] = 1
    df_mc["D_BF_variation_5_up"  ] = 1
    df_mc["D_BF_variation_5_down"] = 1
    df_mc["D_BF_variation_6_up"  ] = 1
    df_mc["D_BF_variation_6_down"] = 1
    df_mc["D_BF_variation_7_up"  ] = 1
    df_mc["D_BF_variation_7_down"] = 1
    df_mc["D_BF_variation_8_up"  ] = 1
    df_mc["D_BF_variation_8_down"] = 1
    df_mc["D_BF_variation_9_up"  ] = 1
    df_mc["D_BF_variation_9_down"] = 1

    def vary_D_BR(df_mc, i, D_BR, selection, debug=False):
        df_mc.loc[selection, f"D_BF_variation_{i}_up"] = 1 / D_BR.n * (D_BR.n + D_BR.s)
        df_mc.loc[selection, f"D_BF_variation_{i}_down"] = 1 / D_BR.n * (D_BR.n - D_BR.s)
        if debug:
            print(
                df_mc.loc[selection, ["__weight_overall__", f"D_BF_variation_{i}_up", f"D_BF_variation_{i}_down"]
                         ].applymap(lambda x: f"{x:4f}").head(n=3)
            )
        return df_mc


    # (321, 211, 999, 999, 999, 999, 999)      57534  # 0
    # (213, 321, 999, 999, 999, 999, 999)      29801  # 1
    # (20213, 321, 999, 999, 999, 999, 999)    22938  # 2
    # (321, 211, 211, 999, 999, 999, 999)      10147  # 3
    # (323, 211, 999, 999, 999, 999, 999)       9623  # 4
    # (321, 211, 211, 211, 999, 999, 999)       7294  # 5
    # (321, 321, 999, 999, 999, 999, 999)       5238  # 6
    # (10311, 211, 999, 999, 999, 999, 999)     4928  # 7
    # (313, 113, 999, 999, 999, 999, 999)       4793  # 8
    # (313, 111, 999, 999, 999, 999, 999)       4746  # 9

    df_mc = vary_D_BR(df_mc, 0,
        ufloat(0.0395, 0.03/100),
        df_mc["sig_d_decay_abs"] == (321, 211, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 1,
        ufloat(0.113, 0.7/100),
        df_mc["sig_d_decay_abs"] == (213, 321, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 2,
        ufloat(0.0750, 0.32/100),
        df_mc["sig_d_decay_abs"] == (20213, 321, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 3,
        ufloat(0.0938 - 0.0104 - 0.0125, 0.16/100),
        df_mc["sig_d_decay_abs"] == (321, 211, 211, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 4,
        ufloat(0.0164 + 0.0231, (0.37e-3**2 + (0.3/100)**2)**0.5),
        df_mc["sig_d_decay_abs"] == (323, 211, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 5,
        ufloat(0.0181, 0.07/100),
        df_mc["sig_d_decay_abs"] == (321, 211, 211, 211, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 6,
        ufloat(0.00408, 0.06e-3),
        df_mc["sig_d_decay_abs"] == (321, 321, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 7,
        ufloat(0.0104, 0.12/100),
        df_mc["sig_d_decay_abs"] == (10311, 211, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 8,
        ufloat(0.0023, 0.05/100),
        df_mc["sig_d_decay_abs"] == (313, 113, 999, 999, 999, 999, 999)
    )

    df_mc = vary_D_BR(df_mc, 9,
        ufloat(0.0195, 0.24/100),
        df_mc["sig_d_decay_abs"] == (313, 111, 999, 999, 999, 999, 999)
    )
"""
