"""
Fit setup info for dedicated R(D) and R(D*) fit.

Felix Metzner 2022
"""


from typing import Tuple, Dict
from uncertainties import ufloat

from templatefitter.plotter.plot_style import KITColors, TangoColors

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.observables.mc_observables import RDStarFitSignalIDwithRecoModeInfo
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollection, MCMatchingCollectionItem

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    ComponentInfo,
    TagEfficiencyCorrectionInfoItem,
    TagEfficiencyCorrectionInfo,
    RecoChannelInfo,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups import rdstar_fit_obs_setups
from rdstar.offline_analysis.fitting.dedicated_fit_approach.utility import generate_sig_id_application_function
from rdstar.offline_analysis.efficiency_calculation.rdstar_efficiency_definitions import (
    get_rdstar_true_charge_sig_id_data_type_mapping,
)

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import (
    FitSetupTriplet,
    base_asimov_data_setup_info,
    base_rdstar_weight_info,
)


__all__ = [
    "rdstar_fit_setup",
    "original__rdstar_fit__component_info_collection",
]


logger = log_to_default(name=__name__)


rdstar_fit_obs_setup = rdstar_fit_obs_setups.FitObsSetupMain


common_rdstar_fit_component_id_column_name: str = RDStarFitSignalIDwithRecoModeInfo.new_column_name
add_common_rdstar_fit_component_id_to = generate_sig_id_application_function(
    sig_col_obs_definition_class=RDStarFitSignalIDwithRecoModeInfo,
    new_col_name=common_rdstar_fit_component_id_column_name,
)


# MC Matching IDs for component definitions

_dstst_l_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.DSSl.id,
    OldMCMatchingIDsQueryInfo.DSSlbkg.id,
)

_bbbar_bkg_sig_ids: Tuple[int, ...] = (
    OldMCMatchingIDsQueryInfo.HBKG.id,
    OldMCMatchingIDsQueryInfo.GBKG.id,
    OldMCMatchingIDsQueryInfo.PiLNu.id,
)

_continuum_bkg_sig_ids: Tuple[int, ...] = (OldMCMatchingIDsQueryInfo.CBKG.id,)

dstst_l_in_BpD_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_dstst_l_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="cb_to_d",
)
dstst_l_in_BzD_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_dstst_l_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="nb_to_d",
)
dstst_l_in_BpDst_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_dstst_l_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="cb_to_dst",
)
dstst_l_in_BzDst_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_dstst_l_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="nb_to_dst",
)

bbbar_in_BpD_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_bbbar_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="cb_to_d",
)
bbbar_in_BzD_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_bbbar_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="nb_to_d",
)
bbbar_in_BpDst_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_bbbar_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="cb_to_dst",
)
bbbar_in_BzDst_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_bbbar_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="nb_to_dst",
)

continuum_in_BpD_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_continuum_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="cb_to_d",
)
continuum_in_BzD_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_continuum_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="nb_to_d",
)
continuum_in_BpDst_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_continuum_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="cb_to_dst",
)
continuum_in_BzDst_bkg_sig_ids: Tuple[int, ...] = RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
    sig_ids=_continuum_bkg_sig_ids,
    is_neutral_signal=None,
    reco_mode="nb_to_dst",
)


# Definition of components:

component_Bp_Dz_tau: ComponentInfo = ComponentInfo(
    name="BpDztau",
    comp_index=10,
    latex_label=LatexDecayProcesses.to_dz_tau_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.Dtau.id,
        is_neutral_signal=False,
        reco_mode=None,
    ),
    color="mediumspringgreen",
)

component_Bz_Dm_tau: ComponentInfo = ComponentInfo(
    name="BzDmtau",
    comp_index=11,
    latex_label=LatexDecayProcesses.to_dm_tau_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.Dtau.id,
        is_neutral_signal=True,
        reco_mode=None,
    ),
    color=KITColors.kit_green,
)


component_Bp_DzS_tau: ComponentInfo = ComponentInfo(
    name="BpDzStau",
    comp_index=20,
    latex_label=LatexDecayProcesses.to_dz_st_tau_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.DStau.id,
        is_neutral_signal=False,
        reco_mode=None,
    ),
    color=TangoColors.chameleon2,
)

component_Bz_DmS_tau: ComponentInfo = ComponentInfo(
    name="BzDmStau",
    comp_index=21,
    latex_label=LatexDecayProcesses.to_dm_st_tau_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.DStau.id,
        is_neutral_signal=True,
        reco_mode=None,
    ),
    color=KITColors.kit_maygreen,
)


component_Bp_Dz_l: ComponentInfo = ComponentInfo(
    name="BpDzl",
    comp_index=30,
    latex_label=LatexDecayProcesses.to_dz_l_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
        is_neutral_signal=False,
        reco_mode=None,
    ),
    color=TangoColors.skyblue1,
)

component_Bz_Dm_l: ComponentInfo = ComponentInfo(
    name="BzDml",
    comp_index=31,
    latex_label=LatexDecayProcesses.to_dm_l_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
        is_neutral_signal=True,
        reco_mode=None,
    ),
    color=KITColors.kit_blue,
)


component_Bp_DzS_l: ComponentInfo = ComponentInfo(
    name="BpDzSl",
    comp_index=40,
    latex_label=LatexDecayProcesses.to_dz_st_l_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
        is_neutral_signal=False,
        reco_mode=None,
    ),
    color="cyan",
)

component_Bz_DmS_l: ComponentInfo = ComponentInfo(
    name="BzDmSl",
    comp_index=41,
    latex_label=LatexDecayProcesses.to_dm_st_l_nu,
    sig_ids=RDStarFitSignalIDwithRecoModeInfo.get_modified_sig_id(
        sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
        is_neutral_signal=True,
        reco_mode=None,
    ),
    color=KITColors.kit_cyan,
)


# D** BKG

component_Dss_in_cB: ComponentInfo = ComponentInfo(
    name="DSS_in_cB",
    comp_index=50,
    latex_label=LatexDecayProcesses.to_dstst_l_nu,
    sig_ids=dstst_l_in_BpD_bkg_sig_ids + dstst_l_in_BpDst_bkg_sig_ids,
    color=KITColors.kit_red,
)
component_Dss_in_nB: ComponentInfo = ComponentInfo(
    name="DSS_in_nB",
    comp_index=51,
    latex_label=LatexDecayProcesses.to_dstst_l_nu,
    sig_ids=dstst_l_in_BzD_bkg_sig_ids + dstst_l_in_BzDst_bkg_sig_ids,
    color=KITColors.kit_red,
)

component_Dss_in_BpD: ComponentInfo = ComponentInfo(
    name="DSS_in_BpD",
    comp_index=50,
    latex_label=LatexDecayProcesses.to_dstst_l_nu,
    sig_ids=dstst_l_in_BpD_bkg_sig_ids,
    color=KITColors.kit_red,
)
component_Dss_in_BzD: ComponentInfo = ComponentInfo(
    name="DSS_in_BzD",
    comp_index=51,
    latex_label=LatexDecayProcesses.to_dstst_l_nu,
    sig_ids=dstst_l_in_BzD_bkg_sig_ids,
    color=KITColors.kit_red,
)
component_Dss_in_BpDst: ComponentInfo = ComponentInfo(
    name="DSS_in_BpDst",
    comp_index=52,
    latex_label=LatexDecayProcesses.to_dstst_l_nu,
    sig_ids=dstst_l_in_BpDst_bkg_sig_ids,
    color=KITColors.kit_red,
)
component_Dss_in_BzDst: ComponentInfo = ComponentInfo(
    name="DSS_in_BzDst",
    comp_index=53,
    latex_label=LatexDecayProcesses.to_dstst_l_nu,
    sig_ids=dstst_l_in_BzDst_bkg_sig_ids,
    color=KITColors.kit_red,
)


# BBbar BKG

component_bbbar_in_cB_bkg: ComponentInfo = ComponentInfo(
    name="BBbarBKG_in_cB",
    comp_index=60,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=bbbar_in_BpD_bkg_sig_ids + bbbar_in_BpDst_bkg_sig_ids,
    color=KITColors.grey,
)
component_bbbar_in_nB_bkg: ComponentInfo = ComponentInfo(
    name="BBbarBKG_in_nB",
    comp_index=61,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=bbbar_in_BzD_bkg_sig_ids + bbbar_in_BzDst_bkg_sig_ids,
    color=KITColors.grey,
)

component_bbbar_in_BpD_bkg: ComponentInfo = ComponentInfo(
    name="BBbarBKG_in_BpD",
    comp_index=60,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=bbbar_in_BpD_bkg_sig_ids,
    color=KITColors.grey,
)
component_bbbar_in_BzD_bkg: ComponentInfo = ComponentInfo(
    name="BBbarBKG_in_BzD",
    comp_index=61,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=bbbar_in_BzD_bkg_sig_ids,
    color=KITColors.grey,
)
component_bbbar_in_BpDst_bkg: ComponentInfo = ComponentInfo(
    name="BBbarBKG_in_BpDst",
    comp_index=62,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=bbbar_in_BpDst_bkg_sig_ids,
    color=KITColors.grey,
)
component_bbbar_in_BzDst_bkg: ComponentInfo = ComponentInfo(
    name="BBbarBKG_in_BzDst",
    comp_index=63,
    latex_label=LatexDecayProcesses.to_b_bbar_bkg,
    sig_ids=bbbar_in_BzDst_bkg_sig_ids,
    color=KITColors.grey,
)


# Continuum BKG

component_continuum_in_cB_bkg: ComponentInfo = ComponentInfo(
    name="CBKG_in_cB",
    comp_index=70,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=continuum_in_BpD_bkg_sig_ids + continuum_in_BpDst_bkg_sig_ids,
    color=KITColors.light_grey,
)
component_continuum_in_nB_bkg: ComponentInfo = ComponentInfo(
    name="CBKG_in_nB",
    comp_index=71,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=continuum_in_BzD_bkg_sig_ids + continuum_in_BzDst_bkg_sig_ids,
    color=KITColors.light_grey,
)

component_continuum_in_BpD_bkg: ComponentInfo = ComponentInfo(
    name="CBKG_in_BpD",
    comp_index=70,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=continuum_in_BpD_bkg_sig_ids,
    color=KITColors.light_grey,
)
component_continuum_in_BzD_bkg: ComponentInfo = ComponentInfo(
    name="CBKG_in_BzD",
    comp_index=71,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=continuum_in_BzD_bkg_sig_ids,
    color=KITColors.light_grey,
)
component_continuum_in_BpDst_bkg: ComponentInfo = ComponentInfo(
    name="CBKG_in_BpDst",
    comp_index=72,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=continuum_in_BpDst_bkg_sig_ids,
    color=KITColors.light_grey,
)
component_continuum_in_BzDst_bkg: ComponentInfo = ComponentInfo(
    name="CBKG_in_BzDst",
    comp_index=73,
    latex_label=LatexDecayProcesses.to_continuum_txt,
    sig_ids=continuum_in_BzDst_bkg_sig_ids,
    color=KITColors.light_grey,
)


rdstar_comps: Tuple[ComponentInfo, ...] = (
    # Signal
    component_Bp_Dz_tau,
    component_Bz_Dm_tau,
    component_Bp_DzS_tau,
    component_Bz_DmS_tau,
    # Normalization
    component_Bp_Dz_l,
    component_Bz_Dm_l,
    component_Bp_DzS_l,
    component_Bz_DmS_l,
    # BKG
    component_Dss_in_cB,
    component_Dss_in_nB,
    # component_Dss_in_BpD,
    # component_Dss_in_BzD,
    # component_Dss_in_BpDst,
    # component_Dss_in_BzDst,
    component_bbbar_in_cB_bkg,
    component_bbbar_in_nB_bkg,
    # component_bbbar_in_BpD_bkg,
    # component_bbbar_in_BzD_bkg,
    # component_bbbar_in_BpDst_bkg,
    # component_bbbar_in_BzDst_bkg,
    component_continuum_in_cB_bkg,
    component_continuum_in_nB_bkg,
    # component_continuum_in_BpD_bkg,
    # component_continuum_in_BzD_bkg,
    # component_continuum_in_BpDst_bkg,
    # component_continuum_in_BzDst_bkg,
)


# RDStar Fit Plot Component Definition
__original_sig_id__rdstar_fit__collection_definition: Tuple[MCMatchingCollectionItem, ...] = tuple(
    [
        MCMatchingCollectionItem(
            name=comp.name,
            sig_ids=comp.sig_ids,
            new_id=comp.comp_index,
            latex_label=comp.latex_label,
            color=comp.color,
        )
        for comp in rdstar_comps
    ]
)
original__rdstar_fit__component_info_collection: MCMatchingCollection = MCMatchingCollection(
    collection_name="original_sig_id__rdstar_fit",
    sig_id_col_name="__tmp__" + common_rdstar_fit_component_id_column_name,
    original_sig_id_col_name=common_rdstar_fit_component_id_column_name,
    collection_definition=__original_sig_id__rdstar_fit__collection_definition,
    base_mc_matching_id_query_infos=OldMCMatchingIDsQueryInfo,
    is_reduced_subset_of_original=False,
)


rdstar_chs: Tuple[RecoChannelInfo, ...] = tuple(
    [
        RecoChannelInfo(
            name="Bz_to_Dm",
            ch_index=0,
            reco_mode_info=RecoModeInfos.neutral_b_to_d_modes,
            components=rdstar_comps,
            observable_infos=rdstar_fit_obs_setup.rdstar_obs_info_Bz_to_Dm,
            tag_eff_corr_factor=ufloat(1.0, 0.05),
        ),
        RecoChannelInfo(
            name="Bm_to_Dz",
            ch_index=1,
            reco_mode_info=RecoModeInfos.charged_b_to_d_modes,
            components=rdstar_comps,
            observable_infos=rdstar_fit_obs_setup.rdstar_obs_info_Bm_to_Dz,
            tag_eff_corr_factor=ufloat(1.0, 0.05),
        ),
        RecoChannelInfo(
            name="Bz_to_Dsm",
            ch_index=2,
            reco_mode_info=RecoModeInfos.neutral_b_to_ds_modes,
            components=rdstar_comps,
            observable_infos=rdstar_fit_obs_setup.rdstar_obs_info_Bz_to_Dsm,
            tag_eff_corr_factor=ufloat(1.0, 0.05),
        ),
        RecoChannelInfo(
            name="Bm_to_Dsz",
            ch_index=3,
            reco_mode_info=RecoModeInfos.charged_b_to_ds_modes,
            components=rdstar_comps,
            observable_infos=rdstar_fit_obs_setup.rdstar_obs_info_Bm_to_Dsz,
            tag_eff_corr_factor=ufloat(1.0, 0.05),
        ),
    ]
)

_rdstar_chs_dict: Dict[str, RecoChannelInfo] = {ch.name: ch for ch in rdstar_chs}


# rdstar_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
#     items=[
#         TagEfficiencyCorrectionInfoItem(key_name="bz_to_d", value=_rdstar_chs_dict["Bz_to_Dm"].tag_eff_corr_factor),
#         TagEfficiencyCorrectionInfoItem(key_name="bpm_to_d", value=_rdstar_chs_dict["Bm_to_Dz"].tag_eff_corr_factor),
#         TagEfficiencyCorrectionInfoItem(key_name="bz_to_ds", value=_rdstar_chs_dict["Bz_to_Dsm"].tag_eff_corr_factor),
#         TagEfficiencyCorrectionInfoItem(key_name="bpm_to_ds", value=_rdstar_chs_dict["Bm_to_Dsz"].tag_eff_corr_factor),
#     ],
#     key_order=("bz_to_d", "bpm_to_d", "bz_to_ds", "bpm_to_ds"),
# )  # This setup was used for initial Asimov tests...


rdstar_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
    items=[
        TagEfficiencyCorrectionInfoItem(key_name="bz_to_d_wwos", value=ufloat(0.76, 0.01)),
        TagEfficiencyCorrectionInfoItem(key_name="bpm_to_d_wwos", value=ufloat(0.82, 0.01)),
        TagEfficiencyCorrectionInfoItem(key_name="bz_to_ds_df", value=ufloat(0.85, 0.11)),
        TagEfficiencyCorrectionInfoItem(key_name="bpm_to_ds_df", value=ufloat(0.87, 0.03)),
    ],
    key_order=("bz_to_d_wwos", "bpm_to_d_wwos", "bz_to_ds_df", "bpm_to_ds_df"),
)


_rdstar_asimov_fit_setup = FitSetupInfoContainer(
    components=rdstar_comps,
    reco_channels=rdstar_chs,
    observable_infos=rdstar_fit_obs_setup.rdstar_obs_infos,
    weight_info=base_rdstar_weight_info,
    data_setup_info=base_asimov_data_setup_info,
    remove_empty_bins=True,
    use_alternative_gap=True,
    tag_efficiency_correction_info=rdstar_tag_eff_corr_info,
    reco_mode_column_name=Cols.decay_mode_column_name,
    component_id_column_name=common_rdstar_fit_component_id_column_name,
    component_id_function=add_common_rdstar_fit_component_id_to,
    sig_id_to_data_type_map_function=get_rdstar_true_charge_sig_id_data_type_mapping,
)


rdstar_fit_setup: FitSetupTriplet = FitSetupTriplet(asimov=_rdstar_asimov_fit_setup)
