"""
Fit setup info for dedicated tagging efficiency calibration fit.

Felix Metzner 2022
"""

from typing import Tuple
from uncertainties import ufloat

from templatefitter.plotter.plot_style import KITColors, TangoColors

from rdstar.utilities.logging import log_to_default

from rdstar.latex_labels.latex_process_labels import LatexDecayProcesses

from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_infos import RecoModeInfos
from rdstar.offline_analysis.mc_matching.old_mc_matching_info import OldMCMatchingIDsQueryInfo
from rdstar.offline_analysis.mc_matching.mc_matching_id_collections import MCMatchingCollection, MCMatchingCollectionItem

from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_setups.fit_setup_utilities import (
    FitSetupTriplet,
    base_asimov_data_setup_info,
    tag_cal_fit_obs_setup,
    base_tag_cal_weight_info,
    tag_cal_bbbar_bkg_sig_ids,
    tag_cal_dstst_l_bkg_sig_ids,
    common_tag_cal_fit_component_id_column_name,
    add_common_tag_cal_fit_component_id_to,
    get_sig_ids_with_charge_info,
)
from rdstar.offline_analysis.fitting.dedicated_fit_approach.fit_info_container import (
    FitSetupInfoContainer,
    ComponentInfo,
    TagEfficiencyCorrectionInfo,
    RecoChannelInfo,
)


__all__ = [
    "tag_cal_fit_setup",
    "tag_cal_fit__component_info_collection",
]


logger = log_to_default(name=__name__)


# Definition of components:

# (Main) D(*) Components
component_Bp_Dz_l_base_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
    reco_type="charged",
    charge_type=None,
    cross_feed_type="wwo_cf",
)
component_Bp_DzS_l_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
    reco_type="charged",
    charge_type=None,
    cross_feed_type="wo_cf",
)

component_Bp_Dwwos_l: ComponentInfo = ComponentInfo(
    name="Bp_to_Dwwos_l",
    comp_index=0,
    latex_label=LatexDecayProcesses.bp_to_dwwostz_l_nu,
    sig_ids=component_Bp_Dz_l_base_ids + component_Bp_DzS_l_ids,
    color=KITColors.kit_blue,
)


component_Bz_Dm_l_base_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.Dl.id,
    reco_type="neutral",
    charge_type=None,
    cross_feed_type="wwo_cf",
)
component_Bz_DmS_l_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
    reco_type="neutral",
    charge_type=None,
    cross_feed_type="wo_cf",
)

component_Bz_Dwwos_l: ComponentInfo = ComponentInfo(
    name="Bz_to_Dwwos_l",
    comp_index=1,
    latex_label=LatexDecayProcesses.bz_to_dwwostm_l_nu,
    sig_ids=component_Bz_Dm_l_base_ids + component_Bz_DmS_l_ids,
    color=KITColors.kit_cyan,
)


# Feed Down D* Components
component_Bp_DzS_l_df: ComponentInfo = ComponentInfo(
    name="BpDzSlDF",
    comp_index=101,
    latex_label=LatexDecayProcesses.to_dz_st_l_nu + " DF",
    sig_ids=get_sig_ids_with_charge_info(
        sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
        reco_type="charged",
        charge_type=None,
        cross_feed_type="cf",
    ),
    color=KITColors.kit_green,
)

component_Bz_DmS_l_df: ComponentInfo = ComponentInfo(
    name="BzDmSlDF",
    comp_index=102,
    latex_label=LatexDecayProcesses.to_dm_st_l_nu + " DF",
    sig_ids=get_sig_ids_with_charge_info(
        sig_ids=OldMCMatchingIDsQueryInfo.DSl.id,
        reco_type="neutral",
        charge_type=None,
        cross_feed_type="cf",
    ),
    color=KITColors.kit_maygreen,
)


# Background Components
tag_cal_charged_b_to_d_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_bbbar_bkg_sig_ids,
    reco_type="c_d",
    charge_type=None,
)
component_charged_b_to_d_recoB_Dss_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_dstst_l_bkg_sig_ids,
    reco_type="c_d",
    charge_type=None,
)
component_charged_b_to_d_reco_continuum_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="c_d",
    charge_type=None,
)

tag_cal_charged_b_to_d_bkg_ids: Tuple[int, ...] = (
    *tag_cal_charged_b_to_d_bbbar_bkg_ids,
    *component_charged_b_to_d_recoB_Dss_ids,
    *component_charged_b_to_d_reco_continuum_bkg_ids,
)


tag_cal_charged_b_to_ds_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_bbbar_bkg_sig_ids,
    reco_type="c_dst",
    charge_type=None,
)
component_charged_b_to_ds_recoB_Dss_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_dstst_l_bkg_sig_ids,
    reco_type="c_dst",
    charge_type=None,
)
component_charged_b_to_ds_reco_continuum_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="c_dst",
    charge_type=None,
)

tag_cal_charged_b_to_ds_bkg_ids: Tuple[int, ...] = (
    *tag_cal_charged_b_to_ds_bbbar_bkg_ids,
    *component_charged_b_to_ds_recoB_Dss_ids,
    *component_charged_b_to_ds_reco_continuum_bkg_ids,
)


component_charged_b_bkg: ComponentInfo = ComponentInfo(
    name="charged_b_BKG",
    comp_index=60,
    latex_label="BKG in " + LatexDecayProcesses.reco_bz_to_d_wwos,
    sig_ids=tag_cal_charged_b_to_d_bkg_ids + tag_cal_charged_b_to_ds_bkg_ids,
    color=TangoColors.aluminium2,
)


tag_cal_neutral_b_to_d_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_bbbar_bkg_sig_ids,
    reco_type="n_d",
    charge_type=None,
)
component_neutral_b_to_d_recoB_Dss_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_dstst_l_bkg_sig_ids,
    reco_type="n_d",
    charge_type=None,
)
component_neutral_b_to_d_reco_continuum_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="n_d",
    charge_type=None,
)

tag_cal_neutral_b_to_d_bkg_ids: Tuple[int, ...] = (
    *tag_cal_neutral_b_to_d_bbbar_bkg_ids,
    *component_neutral_b_to_d_recoB_Dss_ids,
    *component_neutral_b_to_d_reco_continuum_bkg_ids,
)

tag_cal_neutral_b_to_ds_bbbar_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_bbbar_bkg_sig_ids,
    reco_type="n_dst",
    charge_type=None,
)
component_neutral_b_to_ds_recoB_Dss_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=tag_cal_dstst_l_bkg_sig_ids,
    reco_type="n_dst",
    charge_type=None,
)
component_neutral_b_to_ds_reco_continuum_bkg_ids: Tuple[int, ...] = get_sig_ids_with_charge_info(
    sig_ids=OldMCMatchingIDsQueryInfo.CBKG.id,
    reco_type="n_dst",
    charge_type=None,
)

tag_cal_neutral_b_to_ds_bkg_ids: Tuple[int, ...] = (
    *tag_cal_neutral_b_to_ds_bbbar_bkg_ids,
    *component_neutral_b_to_ds_recoB_Dss_ids,
    *component_neutral_b_to_ds_reco_continuum_bkg_ids,
)

component_neutral_b_bkg: ComponentInfo = ComponentInfo(
    name="neutral_b_BKG",
    comp_index=61,
    latex_label="BKG in " + LatexDecayProcesses.reco_bpm_to_d_wwos,
    sig_ids=tag_cal_neutral_b_to_d_bkg_ids + tag_cal_neutral_b_to_ds_bkg_ids,
    color=TangoColors.aluminium4,
)


tag_cal_fit_comps: Tuple[ComponentInfo, ...] = (
    component_Bp_Dwwos_l,
    component_Bz_Dwwos_l,
    component_charged_b_bkg,
    component_neutral_b_bkg,
    component_Bp_DzS_l_df,
    component_Bz_DmS_l_df,
)


# Tag-Cal Fit Plot Component Definition
_plot_tag_cal_fit_comps: Tuple[ComponentInfo, ...] = (
    component_Bp_Dwwos_l,
    component_Bz_Dwwos_l,
    component_Bp_DzS_l_df,
    component_Bz_DmS_l_df,
    component_charged_b_bkg,
    component_neutral_b_bkg,
)

assert set(_plot_tag_cal_fit_comps) == set(tag_cal_fit_comps)

__sig_id__tag_cal_fit__collection_definition: Tuple[MCMatchingCollectionItem, ...] = tuple(
    [
        MCMatchingCollectionItem(
            name=comp.name,
            sig_ids=comp.sig_ids,
            new_id=comp.comp_index,
            latex_label=comp.latex_label,
            color=comp.color,
        )
        for comp in _plot_tag_cal_fit_comps
    ]
)
tag_cal_fit__component_info_collection: MCMatchingCollection = MCMatchingCollection(
    collection_name="original_sig_id__tag_cal_fit",
    sig_id_col_name="__tmp__" + common_tag_cal_fit_component_id_column_name,
    original_sig_id_col_name=common_tag_cal_fit_component_id_column_name,
    collection_definition=__sig_id__tag_cal_fit__collection_definition,
    base_mc_matching_id_query_infos=OldMCMatchingIDsQueryInfo,
    is_reduced_subset_of_original=False,
)


tag_cal_fit_chs: Tuple[RecoChannelInfo, ...] = tuple(
    [
        RecoChannelInfo(
            name="Bz_to_Dm",
            ch_index=0,
            reco_mode_info=RecoModeInfos.neutral_b_to_d_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_fit_obs_setup.rdstar_obs_info_Bz_to_Dm,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="Bm_to_Dz",
            ch_index=1,
            reco_mode_info=RecoModeInfos.charged_b_to_d_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_fit_obs_setup.rdstar_obs_info_Bm_to_Dz,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="Bz_to_Dsm",
            ch_index=2,
            reco_mode_info=RecoModeInfos.neutral_b_to_ds_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_fit_obs_setup.rdstar_obs_info_Bz_to_Dsm,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
        RecoChannelInfo(
            name="Bm_to_Dsz",
            ch_index=3,
            reco_mode_info=RecoModeInfos.charged_b_to_ds_modes,
            components=tag_cal_fit_comps,
            observable_infos=tag_cal_fit_obs_setup.rdstar_obs_info_Bm_to_Dsz,
            tag_eff_corr_factor=ufloat(0.0, 0.00),
        ),
    ]
)


tag_cal_fit_tag_eff_corr_info: TagEfficiencyCorrectionInfo = TagEfficiencyCorrectionInfo(
    items=[],
    key_order=(),
)


_tag_cal_asimov_fit_setup = FitSetupInfoContainer(
    components=tag_cal_fit_comps,
    reco_channels=tag_cal_fit_chs,
    observable_infos=tag_cal_fit_obs_setup.rdstar_obs_infos,
    weight_info=base_tag_cal_weight_info,
    data_setup_info=base_asimov_data_setup_info,
    remove_empty_bins=True,
    use_alternative_gap=True,
    tag_efficiency_correction_info=tag_cal_fit_tag_eff_corr_info,
    reco_mode_column_name=Cols.decay_mode_column_name,
    component_id_column_name=common_tag_cal_fit_component_id_column_name,
    component_id_function=add_common_tag_cal_fit_component_id_to,
)


tag_cal_fit_setup: FitSetupTriplet = FitSetupTriplet(asimov=_tag_cal_asimov_fit_setup)
