"""
Missing Mass Squared Resolution Correction

Felix Metzner 2020
"""

import os
import json
import numpy as np
import pandas as pd

from scipy.special import expit
from scipy import stats as scipy_stats
from typing import Optional, Tuple, List, Dict, NamedTuple, Callable, Any, ClassVar

from rdstar.utilities import PathType
from rdstar.utilities.logging import log_to_default

from rdstar.offline_analysis.mms_resolution_correction.shift_stats import laplace_asymmetric
from rdstar.offline_analysis.mms_resolution_correction.resolution_correction_setups import (
    ResolutionCorrectionSetup,
    ResolutionCorrectionSetupManager,
)


__all__ = [
    "my_lap",
    "ResolutionInfo",
    "MissingMassResolutionCorrector",
]


logger = log_to_default(name=__name__)


class LaplaceSmearingParameters(NamedTuple):
    lp_folding_loc: float = 0.06
    lp_folding_scale: float = 1.0
    lp_sigma_scaling_factor: float = 0.5
    lp_expit_scale: float = -25.0
    lp_expit_asym: float = -0.05
    lp_expit_other: float = 0.0
    min_kappa: float = 1.4
    max_kappa: float = 3.6
    kappa_scale: float = 20.0

    @property
    def kappa_shift(self) -> float:
        return 2.0 * self.kappa_scale * self.lp_folding_loc


my_lap = LaplaceSmearingParameters(
    lp_folding_loc=0.06,
    lp_folding_scale=1.0,
    lp_sigma_scaling_factor=0.5,
    lp_expit_scale=-25.0,
    lp_expit_asym=-0.05,
    lp_expit_other=0.0,
    min_kappa=1.4,
    max_kappa=3.6,
    kappa_scale=20.0,
)


class ResolutionInfo(NamedTuple):
    resolution_correction_setup_name: str
    number_of_bins: int
    sigmas: Tuple[float, ...]
    bin_edges: Optional[Tuple[Tuple[float, float], ...]] = None
    secondary_var: Optional[str] = None
    additional_info: Optional[Dict[str, Any]] = None

    @staticmethod
    def from_file(input_file_path: PathType) -> "ResolutionInfo":
        assert os.path.isfile(input_file_path), input_file_path
        with open(input_file_path, "r") as in_file:
            input_dict = json.load(in_file)

        new_res_info = ResolutionInfo(
            resolution_correction_setup_name=input_dict["resolution_correction_setup_name"],
            number_of_bins=int(input_dict["number_of_bins"]),
            sigmas=tuple(input_dict["sigmas"]),
            bin_edges=tuple([tuple(bin_edge_pair) for bin_edge_pair in input_dict["bin_edges"]]),  # type: ignore
            secondary_var=input_dict["secondary_var"],
            additional_info=input_dict["additional_info"],
        )

        assert len(new_res_info.sigmas) == new_res_info.number_of_bins, (
            new_res_info.number_of_bins,
            len(new_res_info.sigmas),
            new_res_info.sigmas,
        )
        assert all(sig >= 0.0 for sig in new_res_info.sigmas), new_res_info.sigmas
        assert isinstance(new_res_info.bin_edges, tuple), new_res_info.bin_edges
        assert len(new_res_info.bin_edges) == new_res_info.number_of_bins, (
            new_res_info.number_of_bins,
            len(new_res_info.bin_edges),
            new_res_info.bin_edges,
        )
        assert all(isinstance(edges, tuple) for edges in new_res_info.bin_edges), new_res_info.bin_edges
        assert all(len(edges) == 2 for edges in new_res_info.bin_edges), new_res_info.bin_edges
        assert all(edges[0] < edges[1] for edges in new_res_info.bin_edges), new_res_info.bin_edges

        return new_res_info

    @staticmethod
    def create_from_fit_result(
        resolution_correction_setup_name: str,
        secondary_column_name: str,
        number_of_secondary_bins: int,
        secondary_bin_edge_pairs: Tuple[Tuple[float, float], ...],
        mc_results: Dict[int, float],
        data_results: Dict[int, float],
    ) -> "ResolutionInfo":
        assert len(secondary_bin_edge_pairs) == number_of_secondary_bins, (
            number_of_secondary_bins,
            len(secondary_bin_edge_pairs),
            secondary_bin_edge_pairs,
        )

        sigmas = []  # type: List[float]
        additional_info_dict = {}  # type: Dict[str, Any]
        for bin_index, secondary_edges in enumerate(secondary_bin_edge_pairs):
            assert bin_index in mc_results.keys(), bin_index
            assert bin_index in data_results.keys(), bin_index

            mc = mc_results[bin_index]  # type: float
            data = data_results[bin_index]  # type: float
            plain_sigma_diff = abs(data - mc)  # type: float

            if data <= mc:
                diff = 0.0  # type: float
            else:
                diff = np.sqrt(np.abs(data ** 2 - mc ** 2))

            additional_info_dict.update(
                {
                    f"bin_{bin_index}": {
                        "index": bin_index,
                        "lower_bound": secondary_edges[0],
                        "upper_bound": secondary_edges[1],
                        "mc_sigma": mc,
                        "data_sigma": data,
                        "sigma_diff": diff,
                        "plain_sigma_diff": plain_sigma_diff,
                    }
                }
            )

            sigmas.append(diff)

            logger.info(f"Bin {bin_index} ({secondary_edges}): " f"mc={mc}, " f"data={data} " f"diff={diff}")

        assert len(additional_info_dict.keys()) == number_of_secondary_bins
        return ResolutionInfo(
            resolution_correction_setup_name=resolution_correction_setup_name,
            number_of_bins=number_of_secondary_bins,
            sigmas=tuple(sigmas),
            bin_edges=secondary_bin_edge_pairs,
            secondary_var=secondary_column_name,
            additional_info=additional_info_dict,
        )

    def write_to_file(self, file_path: PathType) -> None:
        with open(file_path, "w") as output_file:
            json.dump(self._asdict(), output_file, indent=2)


class MissingMassResolutionCorrector:
    laplace_smearing_appendix: ClassVar[str] = "with_AsymLaplaceSmear"

    def __init__(
        self,
        resolution_correction_setup: ResolutionCorrectionSetup,
        resolution_info_path: Optional[PathType] = None,
        resolution_info: Optional[ResolutionInfo] = None,
    ) -> None:

        self._res_corr_setup = resolution_correction_setup  # type: ResolutionCorrectionSetup

        if isinstance(resolution_info, ResolutionInfo) and resolution_info_path is None:
            self._resolution_info = resolution_info  # type: ResolutionInfo
        elif resolution_info_path is not None and os.path.isfile(resolution_info_path) and resolution_info is None:
            self._resolution_info = ResolutionInfo.from_file(input_file_path=resolution_info_path)
        else:
            res_info = "resolution_info=None" if resolution_info is None else "a ResolutionInfo instance"  # type: str
            raise ValueError(
                f"The resolution info to be used must be provided either"
                f"\tas path via the argument 'resolution_info_path'\nOR\n"
                f"\tdirectly as ResolutionInfo instance via the argument 'resolution_info',\n"
                f"but\n\tresolution_info_path={resolution_info_path}\nAND\n\t{res_info}\nhave been provided!"
            )

        assert self._resolution_info.resolution_correction_setup_name == self._res_corr_setup.name, (
            self._resolution_info.resolution_correction_setup_name,
            self._res_corr_setup.name,
        )

        self._original_col = self._get_original_col_name()  # type: str
        self._new_col = self._get_corrected_col_name()  # type: str

    def add_required_secondary_columns(self, sample: pd.DataFrame) -> None:
        variable_definition_function = self._res_corr_setup.variable_definition  # type: Optional[Callable]
        if variable_definition_function is not None:
            variable_definition_function(df=sample, check_range=False, accept_existing_column=True)

    def ensure_primary_col_available(self, sample: pd.DataFrame) -> None:
        if self._res_corr_setup.primary_col_name not in sample.columns:
            raise RuntimeError(
                f"Required primary column {self._res_corr_setup.primary_col_name} is not available in the data frame!"
            )

    def check_corrected_col_name(
        self,
        expected_col_name: str,
        use_laplace_smeared_version: bool,
    ) -> None:
        if use_laplace_smeared_version:
            col_name = self._get_laplace_smeared_corrected_col_name()  # type: str
        else:
            col_name = self._get_corrected_col_name()
        if not expected_col_name == col_name:
            raise RuntimeError(
                f"Actual new column name '{col_name}' and expected new column name '{expected_col_name}' do not match!"
            )

    def apply_resolution_correction(
        self,
        sample: pd.DataFrame,
        random_seed: int = 13374747,
        correct_overflows: bool = True,
        special_treatment: bool = False,
    ) -> None:
        assert self.original_column in sample.columns, self.original_column
        assert self.secondary_var_col in sample.columns, self.secondary_var_col

        assert self.new_column not in sample.columns, self.new_column

        np_rand_state = np.random.RandomState(seed=int(random_seed + 0.01 * len(sample.index)))

        sample.loc[:, self.new_column] = np.nan

        asymmetric_laplace_smeared_col_name = "None__This_should_not_happen"  # type: str
        if special_treatment:
            asymmetric_laplace_smeared_col_name = self._get_laplace_smeared_corrected_col_name()
            assert asymmetric_laplace_smeared_col_name not in sample.columns, asymmetric_laplace_smeared_col_name
            sample.loc[:, asymmetric_laplace_smeared_col_name] = np.nan

        _applied_conditions = []  # type: List[pd.Series]
        for bin_index, (sigma, bin_edges) in enumerate(zip(self.sigmas, self.bin_edge_pairs)):
            if correct_overflows and (bin_index == 0 or bin_index == self.number_of_bins - 1):
                if self.number_of_bins == 1:
                    condition = pd.Series([True] * len(sample.index))  # type: pd.Series
                    assert len(condition.index) == len(sample.index)
                elif bin_index == 0:
                    condition = sample[self.secondary_var_col] < bin_edges[1]
                elif bin_index == self.number_of_bins - 1:
                    condition = sample[self.secondary_var_col] >= bin_edges[0]
                else:
                    raise RuntimeError("This case should not be reached...")
            else:
                left_condition = sample[self.secondary_var_col] >= bin_edges[0]

                if bin_index == self.number_of_bins - 1:
                    # Special case for last bin
                    right_condition = sample[self.secondary_var_col] <= bin_edges[1]
                else:
                    right_condition = sample[self.secondary_var_col] < bin_edges[1]

                condition = left_condition & right_condition

            assert all(not np.any(condition & c) for c in _applied_conditions), (
                self.new_column,
                self._res_corr_setup.name,
                bin_index,
                bin_edges,
            )
            _applied_conditions.append(condition)

            assert pd.isnull(sample.loc[condition, self.new_column]).all(), (self.new_column, self._res_corr_setup.name)

            if sigma == 0.0:
                original_values = sample.loc[condition, self.original_column].values  # type: np.ndarray
                sample.loc[condition, self.new_column] = original_values
                if special_treatment:
                    sample.loc[condition, asymmetric_laplace_smeared_col_name] = original_values
            else:
                sample.loc[condition, self.new_column] = np_rand_state.normal(
                    loc=sample.loc[condition, self.original_column].values,
                    scale=sigma,
                )
                if special_treatment:
                    input_values = sample.loc[condition, self.original_column].values  # type: np.ndarray
                    asymmetric_laplace_smeared = self.calculate_asymmetrically_smeared_values_for_sigma(
                        input_values=input_values,
                        sigma_value=sigma,
                        np_random_state_instance=np_rand_state,
                    )  # type: np.ndarray

                    sample.loc[condition, asymmetric_laplace_smeared_col_name] = asymmetric_laplace_smeared

        if not correct_overflows:
            left_condition = sample[self.secondary_var_col] < self.bin_edge_pairs[0][0]
            right_condition = sample[self.secondary_var_col] > self.bin_edge_pairs[-1][1]
            condition = left_condition & right_condition
            assert pd.isnull(sample.loc[condition, self.new_column]).all()
            sample.loc[condition, self.new_column] = sample.loc[condition, self.original_column]

            if special_treatment:
                assert pd.isnull(sample.loc[condition, asymmetric_laplace_smeared_col_name]).all()
                sample.loc[condition, asymmetric_laplace_smeared_col_name] = sample.loc[condition, self.original_column]

        assert not pd.isnull(sample[self.new_column]).any()
        if special_treatment:
            assert not pd.isnull(sample[asymmetric_laplace_smeared_col_name]).any()

    @staticmethod
    def calculate_asymmetrically_smeared_values_for_sigma(
        input_values: np.ndarray,
        sigma_value: float,
        np_random_state_instance: np.random.RandomState,
    ) -> np.ndarray:

        lp_folding_norm = 1.0 / scipy_stats.norm.pdf(
            0.0,
            loc=my_lap.lp_folding_loc,
            scale=my_lap.lp_folding_scale,
        )  # type: float
        lp_folded_sigmas = (
            my_lap.lp_expit_other + expit(my_lap.lp_expit_scale * (input_values + my_lap.lp_expit_asym))
        ) * (
            sigma_value
            * my_lap.lp_sigma_scaling_factor
            * lp_folding_norm
            * scipy_stats.norm.pdf(
                input_values,
                loc=my_lap.lp_folding_loc,
                scale=my_lap.lp_folding_scale,
            )
        )  # type: np.ndarray

        kappa_values = my_lap.min_kappa + (my_lap.max_kappa - my_lap.min_kappa) / (
            1.0 + np.exp(my_lap.kappa_shift + my_lap.kappa_scale * input_values)
        )  # type: np.ndarray

        asymmetric_laplace_smeared = laplace_asymmetric.rvs(
            loc=input_values,
            kappa=kappa_values,
            scale=lp_folded_sigmas,
            random_state=np_random_state_instance,
        )  # type: np.ndarray
        return asymmetric_laplace_smeared

    def add_corrected_column_for_data(
        self,
        sample: pd.DataFrame,
        special_treatment: bool = False,
    ) -> None:
        orig_column = self._get_original_col_name()  # type: str
        new_column = self._get_corrected_col_name()  # type: str

        assert orig_column in sample.columns, orig_column
        assert new_column not in sample.columns, new_column

        sample.loc[:, new_column] = sample[orig_column].values
        assert new_column in sample.columns, new_column

        if special_treatment:
            asymmetric_laplace_smeared_col_name = self._get_laplace_smeared_corrected_col_name()  # type: str
            assert asymmetric_laplace_smeared_col_name not in sample.columns, asymmetric_laplace_smeared_col_name
            sample.loc[:, asymmetric_laplace_smeared_col_name] = sample[orig_column].values

    def add_corrected_column_to_empty_data_frame(
        self,
        sample: pd.DataFrame,
        special_treatment: bool = False,
    ) -> None:
        orig_column = self._get_original_col_name()  # type: str
        new_column = self._get_corrected_col_name()  # type: str
        assert orig_column in sample.columns, orig_column
        assert new_column not in sample.columns, new_column

        sample = sample.assign(**{new_column: sample[orig_column]})
        assert new_column in sample.columns, new_column

        if special_treatment:
            asymmetric_laplace_smeared_col_name = self._get_laplace_smeared_corrected_col_name()  # type: str
            assert asymmetric_laplace_smeared_col_name not in sample.columns, asymmetric_laplace_smeared_col_name
            sample = sample.assign(**{asymmetric_laplace_smeared_col_name: sample[orig_column]})
            assert asymmetric_laplace_smeared_col_name in sample.columns, asymmetric_laplace_smeared_col_name

    @property
    def original_column(self) -> str:
        return self._original_col

    @property
    def new_column(self) -> str:
        return self._new_col

    @property
    def secondary_var_col(self) -> Optional[str]:
        return self._resolution_info.secondary_var

    @property
    def bin_edge_pairs(self) -> Tuple[Tuple[float, float], ...]:
        assert isinstance(self._resolution_info.bin_edges, tuple), self._resolution_info.bin_edges
        return self._resolution_info.bin_edges

    @property
    def sigmas(self) -> Tuple[float, ...]:
        return self._resolution_info.sigmas

    @property
    def number_of_bins(self) -> int:
        return self._resolution_info.number_of_bins

    def _get_original_col_name(self) -> str:
        return self._res_corr_setup.primary_col_name

    def _get_corrected_col_name(self) -> str:
        return self._res_corr_setup.get_result_col_name(appendix=None)

    def _get_laplace_smeared_corrected_col_name(self) -> str:
        return self._res_corr_setup.get_result_col_name(appendix=self.laplace_smearing_appendix)

    @staticmethod
    def get_single_stage_original_col_name(correction_setup_name: str) -> str:
        # This version is only for single stage resolution correction applications
        # for which no changes to the setup are made!
        cor_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=correction_setup_name,
        )  # type: ResolutionCorrectionSetup
        return cor_setup.primary_col_name

    @staticmethod
    def get_single_stage_corrected_col_name(
        correction_setup_name: str,
        use_laplace_smeared_version: bool,
    ) -> str:
        # This version is only for single stage resolution correction applications
        # for which no changes to the setup are made!
        cor_setup = ResolutionCorrectionSetupManager.get_setup_definition(
            name=correction_setup_name
        )  # type: ResolutionCorrectionSetup
        appendix = MissingMassResolutionCorrector.laplace_smearing_appendix if use_laplace_smeared_version else None
        return cor_setup.get_result_col_name(appendix=appendix)

    @staticmethod
    def get_main_single_stage_correction_column_name(use_laplace_smeared_version: bool) -> str:
        # This version is only for single stage resolution correction applications
        # for which no changes to the setup are made!
        main_res_corr_setup_name = ResolutionCorrectionSetupManager.setup_name_to_use  # type: str
        return MissingMassResolutionCorrector.get_single_stage_corrected_col_name(
            correction_setup_name=main_res_corr_setup_name,
            use_laplace_smeared_version=use_laplace_smeared_version,
        )
