"""
Additional plotting tools for validation of observables.

Felix Metzner 2021
"""

import os
import pandas as pd

from matplotlib import pyplot as plt
from typing import Optional, Union, List, Tuple, Dict

from templatefitter.binned_distributions.binning import Binning
from templatefitter.plotter.histogram_variable import HistVariable
from templatefitter.plotter.histogram_plots import SimpleHistogramPlot
from templatefitter.plotter.plot_utilities import export, FigureType, AxesType
from templatefitter.plotter.plot_style import KITColors, set_matplotlibrc_params
from templatefitter.plotter.validation_plots import BinCompositionPlot, BinMigrationPlot

from rdstar.utilities import PathType

from rdstar.offline_analysis.rdstar_plots import rdstar_color_map
from rdstar.offline_analysis.observables.common_observables import Cols
from rdstar.offline_analysis.reco_modes.reco_mode_collections import BSigRecoModes
from rdstar.offline_analysis.reco_modes.reco_mode_utilities import RecoModeDictType, RecoModeGroupNameToTitleDictType


__all__ = [
    "plot_bin_migration",
    "plot_bin_composition",
]

default_rdstar_reco_modes: RecoModeDictType = {k: v for k, v in BSigRecoModes.main_modes.main_reco_mode_infos_as_tuple}
default_rdstar_reco_mode_labels: RecoModeGroupNameToTitleDictType = (
    BSigRecoModes.main_modes.get_reco_mode_name_to_latex_label_mapping(short=False)
)


def plot_bin_migration(
    mc_df: pd.DataFrame,
    from_hist_var: HistVariable,
    to_hist_var: HistVariable,
    from_to_label_appendix: Optional[Tuple[Optional[str], Optional[str]]] = None,
    weight_col: Optional[str] = None,
    normalize_to_origin: bool = True,
    dec_modes: Optional[RecoModeDictType] = None,
    dec_titles: Optional[RecoModeGroupNameToTitleDictType] = None,
    file_name_suffix: Optional[str] = None,
    show_color_bar: bool = False,
    fig_size: Tuple[float, float] = (12.0, 10.0),
    export_path: Optional[Union[str, PathType]] = None,
    dry_run: bool = False,
) -> Dict[str, PathType]:

    file_name_end: str = "" if file_name_suffix is None else f"_{file_name_suffix}"
    var_info_for_file: str = f"{from_hist_var.df_label}_to_{to_hist_var.df_label}"
    file_name_base: str = f"bin_migration_plot_{var_info_for_file}{file_name_end}_"

    _dec_modes: RecoModeDictType = default_rdstar_reco_modes if dec_modes is None else dec_modes
    _dec_titles: RecoModeGroupNameToTitleDictType = default_rdstar_reco_mode_labels if dec_titles is None else dec_titles

    output_dict: Dict[str, PathType] = {}

    all_mode_ids: Tuple[int, ...] = tuple([d_id for d_ids in _dec_modes.values() for d_id in d_ids])
    mode_infos: List[Tuple[str, Tuple[int, ...]]] = [(k, tuple(v)) for k, v in _dec_modes.items()]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        file_name: str = f"{file_name_base}{mode_name}"

        if export_path is not None:
            output_path: PathType = os.path.join(export_path, f"{file_name}.pdf")
            assert file_name not in output_dict.keys(), file_name
            assert output_path not in output_dict.values(), (file_name, output_path)
            output_dict.update({file_name: output_path})

        if dry_run:
            continue

        migration_plot = BinMigrationPlot(
            from_variable=from_hist_var,
            to_variable=to_hist_var,
            weight_column=weight_col,
            color_map=rdstar_color_map,
            from_to_label_appendix=from_to_label_appendix,
            max_number_of_ticks=32,
            max_bins_for_labels=30,
        )

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=fig_size)  # type: FigureType, AxesType

        migration_plot.plot_on(
            df=mc_df.loc[mc_df[Cols.decay_mode_column_name].isin(mode_ids)],
            ax=ax,
            normalize_to_origin=normalize_to_origin,
            show_color_bar=show_color_bar,
        )

        ax.set_title(_dec_titles[mode_name], loc="right")

        if export_path is not None:
            export(fig=fig, filename=file_name, target_dir=export_path, close_figure=True)

    return output_dict


def plot_bin_composition(
    mc_df: pd.DataFrame,
    main_hist_var: HistVariable,
    secondary_hist_var: HistVariable,
    secondary_binning: Binning,
    secondary_hist_var_label_mapping: Optional[Dict[int, str]] = None,
    weight_col: Optional[str] = None,
    plot_shape: bool = False,
    dec_modes: Optional[RecoModeDictType] = None,
    dec_titles: Optional[RecoModeGroupNameToTitleDictType] = None,
    file_name_suffix: Optional[str] = None,
    draw_legend: bool = True,
    legend_columns: int = 1,
    fig_size: Tuple[float, float] = (8.0, 5.0),
    export_path: Optional[Union[str, PathType]] = None,
    dry_run: bool = False,
) -> Dict[str, PathType]:

    _dec_modes: RecoModeDictType = default_rdstar_reco_modes if dec_modes is None else dec_modes
    _dec_titles: RecoModeGroupNameToTitleDictType = default_rdstar_reco_mode_labels if dec_titles is None else dec_titles

    file_name_end: str = "" if file_name_suffix is None else f"_{file_name_suffix}"
    var_info_for_file: str = f"{main_hist_var.df_label}_for_{secondary_hist_var.df_label}"
    file_name_base: str = f"bin_composition_plot_{var_info_for_file}{file_name_end}_"

    output_dict: Dict[str, PathType] = {}

    all_mode_ids: Tuple[int, ...] = tuple([d_id for d_ids in _dec_modes.values() for d_id in d_ids])
    mode_infos: List[Tuple[str, Tuple[int, ...]]] = [(k, tuple(v)) for k, v in _dec_modes.items()]
    mode_infos.append(("all_modes", all_mode_ids))

    for mode_name, mode_ids in mode_infos:
        file_name: str = f"{file_name_base}{mode_name}"

        if export_path is not None:
            output_path: PathType = os.path.join(export_path, f"{file_name}.pdf")
            assert file_name not in output_dict.keys(), file_name
            assert output_path not in output_dict.values(), (file_name, output_path)
            output_dict.update({file_name: output_path})

        if dry_run:
            continue

        set_matplotlibrc_params()

        bc_plot = BinCompositionPlot(
            variable=main_hist_var,
            secondary_variable=secondary_hist_var,
            secondary_variable_binning=secondary_binning,
            weight_column=weight_col,
            secondary_tick_label_mapping=secondary_hist_var_label_mapping,
        )

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=fig_size)  # type: FigureType, AxesType

        bc_plot.plot_on(
            df=mc_df.loc[mc_df[Cols.decay_mode_column_name].isin(mode_ids)],
            ax=ax,
            draw_legend=draw_legend,
            legend_columns=legend_columns,
        )

        if plot_shape:
            shape_plot = SimpleHistogramPlot(variable=main_hist_var)
            shape_plot.add_component(
                label="Shape",
                data=mc_df,
                weights=weight_col,
                hist_type="step",
                color=KITColors.light_grey,
            )

            ax2: AxesType = ax.twinx()

            shape_plot.plot_on(
                ax=ax2,
                draw_legend=False,
                normed=True,
            )
            ax2.set_ylabel(ylabel="")
            ax2.set_yticks(ticks=[])
            ax2.set_yticklabels(labels=[])

        ax.set_title(_dec_titles[mode_name], loc="right")

        if export_path is not None:
            export(fig=fig, filename=file_name, target_dir=export_path, close_figure=True)

    return output_dict
